#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Main Debug Authentication Tool application."""

import contextlib
import logging
import os
import struct
import sys
from dataclasses import dataclass
from time import sleep
from typing import Callable, Iterator, List, Optional

import click
import colorama

from spsdk import SPSDK_DATA_FOLDER, SPSDKError
from spsdk.apps.blhost_helper import progress_bar
from spsdk.apps.elftosb_utils.sb_31_helper import RootOfTrustInfo
from spsdk.apps.utils import spsdk_logger
from spsdk.apps.utils.common_cli_options import (
    CommandsTreeGroupAliasedGetCfgTemplate,
    spsdk_apps_common_options,
    spsdk_plugin_option,
)
from spsdk.apps.utils.utils import (
    INT,
    SPSDKAppError,
    catch_spsdk_error,
    check_file_exists,
    format_raw_data,
    parse_file_and_size,
    parse_hex_data,
)
from spsdk.dat import DebugAuthenticateResponse, DebugAuthenticationChallenge, dm_commands
from spsdk.dat.debug_credential import DebugCredential
from spsdk.dat.debug_mailbox import DebugMailbox
from spsdk.debuggers.utils import PROBES, open_debug_probe, test_ahb_access
from spsdk.exceptions import SPSDKValueError
from spsdk.utils.images import BinaryImage
from spsdk.utils.misc import find_file, load_binary, load_configuration, load_text, write_file
from spsdk.utils.plugins import load_plugin_from_source

logger = logging.getLogger(__name__)
NXPDEBUGMBOX_DATA_FOLDER: str = os.path.join(SPSDK_DATA_FOLDER, "nxpdebugmbox")


def get_debug_probe_options_help() -> str:
    """Get Click help for debug probe user params.

    :return: Help string.
    """
    ret = (
        "This option could be used multiply to setup non-standard option for debug probe.\n\n"
        "The example of use: -o KEY=VALUE"
    )
    for probe, probe_cls in PROBES.items():
        options_help = probe_cls.get_options_help()
        if options_help:
            ret += f"\n\n[{probe}]:"
            for option, help_text in options_help.items():
                ret += f"\n\n  [{option}]: {help_text}"

    return ret


@dataclass
class DatProtocol:
    """Debug Authentication protocol."""

    VERSIONS = [
        "1.0",
        "1.1",
        "2.0",
        "2.1",
        "2.2",
    ]

    version: str

    def is_rsa(self) -> bool:
        """Determine whether rsa or ecc is used.

        :return: True if the protocol is RSA type. False otherwise
        """
        protocol_version = self.version.split(".")
        is_rsa = protocol_version[0] == "1"
        return is_rsa

    def validate(self) -> None:
        """Validate protocol value.

        :raises SPSDKValueError: In case that protocol is using unsupported key type.
        """
        if self.version not in self.VERSIONS:
            raise SPSDKValueError(f"Unsupported protocol '{self.version}' was given.")


@dataclass
class DebugProbeParams:
    """Debug probe related parameters."""

    interface: str
    serial_no: str
    debug_probe_user_params: dict


@dataclass
class DebugMailboxParams:
    """Debug mailbox related parameters."""

    reset: bool
    more_delay: float
    operation_timeout: int


@contextlib.contextmanager
def _open_debugmbox(
    debug_probe_params: DebugProbeParams,
    debug_mailbox_params: DebugMailboxParams,
) -> Iterator[DebugMailbox]:
    """Method opens DebugMailbox object based on input arguments.

    :param debug_probe_params: DebugProbeParams object holding information about parameters for debug probe.
    :param debug_mailbox_params: DebugMailboxParams object holding information about parameters for debug mailbox.
    :return: Active DebugMailbox object.
    :raises SPSDKError: Raised with any kind of problems with debug probe.
    """
    with open_debug_probe(
        interface=debug_probe_params.interface,
        serial_no=debug_probe_params.serial_no,
        debug_probe_params=debug_probe_params.debug_probe_user_params,
    ) as debug_probe:
        dm = DebugMailbox(
            debug_probe=debug_probe,
            reset=debug_mailbox_params.reset,
            moredelay=debug_mailbox_params.more_delay,
            op_timeout=debug_mailbox_params.operation_timeout,
        )
        try:
            yield dm
        except SPSDKError as exc:
            raise SPSDKError("Problem with debug probe occurred") from exc
        finally:
            dm.close()


@click.group(name="nxpdebugmbox", no_args_is_help=True, cls=CommandsTreeGroupAliasedGetCfgTemplate)
@click.option(
    "-i",
    "--interface",
    type=click.Choice(list(PROBES.keys())),
    help="Probe interface selection, if not specified, all available debug probe interfaces are used.",
)
@click.option(
    "-s",
    "--serial-no",
    help="Debug probe hardware ID/serial number to select the probe in system.",
)
@click.option(
    "-p",
    "--protocol",
    "protocol",
    metavar="VERSION",
    help=f"Set the protocol version. Currently this option is used for gendc and auth sub commands"
    f'Available options are: {", ".join(DatProtocol.VERSIONS)}',
    type=click.Choice(DatProtocol.VERSIONS),
)
@click.option(
    "-t",
    "--timing",
    type=float,
    default=0.0,
    help="Duration of additional delay after reset sequence, defaults to 0 seconds",
)
@click.option(
    "-n",
    "--no-reset",
    "no_reset",
    is_flag=True,
    default=True,
    help=(
        "Omit reset of debug mailbox during initialization,"
        " default behavior is reset debug mailbox during initialization."
    ),
)
@click.option(
    "-o",
    "--debug-probe-option",
    multiple=True,
    help=get_debug_probe_options_help(),
)
@click.option(
    "--operation-timeout",
    type=INT(),
    default="1000",
    help="Special option to change the standard operation timeout used"
    " for communication with debug mailbox. Default value is 1000ms.",
)
@spsdk_apps_common_options
@click.pass_context
def main(
    ctx: click.Context,
    interface: str,
    protocol: str,
    log_level: int,
    timing: float,
    serial_no: str,
    debug_probe_option: List[str],
    no_reset: bool,
    operation_timeout: int,
) -> int:
    """Tool for working with Debug Mailbox."""
    spsdk_logger.install(level=log_level)

    probe_user_params = {}
    for par in debug_probe_option:
        if par.count("=") != 1:
            raise SPSDKError(f"Invalid -o parameter {par}!")

        par_splitted = par.split("=")
        probe_user_params[par_splitted[0]] = par_splitted[1]

    ctx.obj = {
        "debug_mailbox_params": DebugMailboxParams(
            reset=no_reset, more_delay=timing, operation_timeout=operation_timeout
        ),
        "debug_probe_params": DebugProbeParams(
            interface=interface, serial_no=serial_no, debug_probe_user_params=probe_user_params
        ),
        "protocol": DatProtocol(version=protocol),
    }

    return 0


@main.command(name="auth", no_args_is_help=True)
@click.option("-b", "--beacon", type=INT(), help="Authentication beacon")
@click.option("-c", "--certificate", help="Path to Debug Credentials.")
@click.option("-k", "--key", help="Path to DCK private key.")
@click.option(
    "-n",
    "--no-exit",
    is_flag=True,
    help="When used, exit debug mailbox command is not executed after debug authentication.",
)
@click.option(
    "-x",
    "--nxp-keys",
    type=bool,
    is_flag=True,
    default=False,
    help="Use the ROM NXP keys to authenticate.",
)
@click.pass_obj
def auth_command(
    pass_obj: dict, beacon: int, certificate: str, key: str, no_exit: bool, nxp_keys: bool
) -> None:
    """Perform the Debug Authentication.

    The -p protocol option must be defined in main application.
    """
    auth(
        pass_obj["debug_probe_params"],
        pass_obj["debug_mailbox_params"],
        pass_obj["protocol"],
        beacon,
        certificate,
        key,
        no_exit,
        nxp_keys,
    )


def auth(
    debug_probe_params: DebugProbeParams,
    debug_mailbox_params: DebugMailboxParams,
    protocol: DatProtocol,
    beacon: int,
    certificate: str,
    key: str,
    no_exit: bool,
    nxp_keys: bool,
) -> None:
    """Perform the Debug Authentication.

    :param debug_probe_params: DebugProbeParams object holding information about parameters for debug probe.
    :param debug_mailbox_params: DebugMailboxParams object holding information about parameters for debug mailbox.
    :param protocol: Debug authentication protocol.
    :param beacon: Authentication beacon.
    :param certificate: Path to Debug Credentials.
    :param key: Path to DCK private key.
    :param no_exit: When true, exit debug mailbox command is not executed after debug authentication.
    :param nxp_keys: When true, Use the ROM NXP keys to authenticate.
    :raises SPSDKAppError: Raised if any error occurred.
    """
    protocol.validate()
    try:
        logger.info("Starting Debug Authentication")

        with _open_debugmbox(debug_probe_params, debug_mailbox_params) as mail_box:
            debug_cred_data = load_binary(certificate)
            debug_cred = DebugCredential.parse(debug_cred_data)
            dac_rsp_len = 30 if debug_cred.HASH_LENGTH == 48 and debug_cred.socc in [4, 6] else 26
            if nxp_keys:
                dac_data = dm_commands.NxpDebugAuthenticationStart(
                    dm=mail_box, resplen=dac_rsp_len
                ).run()
            else:
                dac_data = dm_commands.DebugAuthenticationStart(
                    dm=mail_box, resplen=dac_rsp_len
                ).run()
            # convert List[int] to bytes
            dac_data_bytes = struct.pack(f"<{len(dac_data)}I", *dac_data)
            dac = DebugAuthenticationChallenge.parse(dac_data_bytes)
            logger.info(f"DAC: \n{dac.info()}")
            dac.validate_against_dc(debug_cred)
            dar = DebugAuthenticateResponse.create(
                version=protocol.version,
                dc=debug_cred,
                auth_beacon=beacon,
                dac=dac,
                dck=key,
            )
            logger.info(f"DAR:\n{dar.info()}")
            dar_data = dar.export()
            # convert bytes to List[int]
            dar_data_words = list(struct.unpack(f"<{len(dar_data) // 4}I", dar_data))
            if nxp_keys:
                dar_response = dm_commands.NxpDebugAuthenticationResponse(
                    dm=mail_box, paramlen=len(dar_data_words)
                ).run(dar_data_words)
            else:
                dar_response = dm_commands.DebugAuthenticationResponse(
                    dm=mail_box, paramlen=len(dar_data_words)
                ).run(dar_data_words)
            logger.debug(f"DAR response: {dar_response}")
            if not no_exit:
                exit_response = dm_commands.ExitDebugMailbox(dm=mail_box).run()
                logger.debug(f"Exit response: {exit_response}")
                # Re-open debug probe
                mail_box.debug_probe.close()
                mail_box.debug_probe.open()
                # Do test of access to AHB bus
                sleep(0.2)
                ahb_access_granted = test_ahb_access(mail_box.debug_probe)
                res_str = (
                    (colorama.Fore.GREEN + "successfully")
                    if ahb_access_granted
                    else (colorama.Fore.RED + "without AHB access")
                )
                logger.info(f"Debug Authentication ends {res_str}{colorama.Fore.RESET}.")
                if not ahb_access_granted:
                    raise SPSDKAppError()
            else:
                logger.info(
                    "Debug Authentication ends without exit and without test of AHB access."
                )

    except SPSDKError as e:
        raise SPSDKAppError(
            f"{colorama.Fore.RED}Debug Mailbox authentication failed:{colorama.Fore.RESET}\n{e}"
        ) from e


@main.command(name="reset")
@click.option(
    "-h",
    "--hard-reset",
    is_flag=True,
    default=False,
    help="When used, the hardware reset is used instead of debug mailbox reset.",
)
@click.pass_obj
def reset_command(pass_obj: dict, hard_reset: bool) -> None:
    """Reset MCU by DebugMailBox.

    The reset command implemented in NXPDEBUGMBOX has two modes (option -h):

    Reset by RESET REQUEST of debug mailbox that causes the reset of MCU by SYSRESET_REQ.
    The chip is reset, but the ROM code returns back the chip into debug mailbox handler
    (without -h/--hard-reset option).

    Reset by external reset signal. This reset is done by asserting external reset signal over
    debug probe. After this reset type the chip behavior is same as after standard reset button on the board.
    (with -h/--hard-reset option)
    """
    reset(pass_obj["debug_probe_params"], pass_obj["debug_mailbox_params"], hard_reset)
    click.echo("Reset MCU by Debug Mailbox succeeded.")


def reset(
    debug_probe_params: DebugProbeParams, debug_mailbox_params: DebugMailboxParams, hard_reset: bool
) -> None:
    """Reset MCU by DebugMailBox.

    :param debug_probe_params: DebugProbeParams object holding information about parameters for debug probe.
    :param debug_mailbox_params: DebugMailboxParams object holding information about parameters for debugmailbox.
    :param hard_reset: If true, use the hardware reset instead of debug mailbox reset.
    :raises SPSDKAppError: Raised if any error occurred.
    """
    try:
        debug_mailbox_params.reset = True
        if hard_reset:
            with open_debug_probe(
                interface=debug_probe_params.interface,
                serial_no=debug_probe_params.serial_no,
                debug_probe_params=debug_probe_params.debug_probe_user_params,
            ) as debug_probe:
                debug_probe.reset()
        else:
            with _open_debugmbox(debug_probe_params, debug_mailbox_params):
                pass
    except Exception as e:
        raise SPSDKAppError(f"Reset MCU by Debug Mailbox failed: {e}") from e


@main.command(name="start")
@click.pass_obj
def start_command(pass_obj: dict) -> None:
    """Start DebugMailBox."""
    start(pass_obj["debug_probe_params"], pass_obj["debug_mailbox_params"])
    click.echo("Start Debug Mailbox succeeded")


def start(debug_probe_params: DebugProbeParams, debug_mailbox_params: DebugMailboxParams) -> None:
    """Start DebugMailBox.

    :param debug_probe_params: DebugProbeParams object holding information about parameters for debug probe.
    :param debug_mailbox_params: DebugMailboxParams object holding information about parameters for debugmailbox.
    :raises SPSDKAppError: Raised if any error occurred.
    """
    try:
        with _open_debugmbox(debug_probe_params, debug_mailbox_params) as mail_box:
            dm_commands.StartDebugMailbox(dm=mail_box).run()
    except Exception as e:
        raise SPSDKAppError(f"Start Debug Mailbox failed: {e}") from e


@main.command(name="exit")
@click.pass_obj
def exit_command(pass_obj: dict) -> None:  # pylint: disable=redefined-builtin
    """Exit DebugMailBox."""
    exit_debug_mbox(pass_obj["debug_probe_params"], pass_obj["debug_mailbox_params"])
    click.echo("Exit Debug Mailbox succeeded")


def exit_debug_mbox(
    debug_probe_params: DebugProbeParams, debug_mailbox_params: DebugMailboxParams
) -> None:  # pylint: disable=redefined-builtin
    """Exit DebugMailBox.

    :param debug_probe_params: DebugProbeParams object holding information about parameters for debug probe.
    :param debug_mailbox_params: DebugMailboxParams object holding information about parameters for debugmailbox.
    :raises SPSDKAppError: Raised if any error occurred.
    """
    try:
        with _open_debugmbox(debug_probe_params, debug_mailbox_params) as mail_box:
            dm_commands.ExitDebugMailbox(dm=mail_box).run()
    except Exception as e:
        raise SPSDKAppError(f"Exit Debug Mailbox failed: {e}") from e


@main.command(name="erase")
@click.pass_obj
def erase_command(pass_obj: dict) -> None:
    """Erase Flash."""
    erase(pass_obj["debug_probe_params"], pass_obj["debug_mailbox_params"])
    click.echo("Mass flash erase succeeded")


def erase(debug_probe_params: DebugProbeParams, debug_mailbox_params: DebugMailboxParams) -> None:
    """Erase Flash.

    :param debug_probe_params: DebugProbeParams object holding information about parameters for debug probe.
    :param debug_mailbox_params: DebugMailboxParams object holding information about parameters for debugmailbox.
    :raises SPSDKAppError: Raised if any error occurred.
    """
    try:
        with _open_debugmbox(debug_probe_params, debug_mailbox_params) as mail_box:
            dm_commands.EraseFlash(dm=mail_box).run()
    except Exception as e:
        raise SPSDKAppError(f"Mass flash erase failed: {e}") from e


@main.command(name="famode", no_args_is_help=True)
@click.option(
    "-m",
    "--message",
    type=click.Path(),
    required=False,
    help="Path to message file.",
)
@click.pass_obj
def famode_command(pass_obj: dict, message: str) -> None:
    """Set Fault Analysis Mode."""
    famode(pass_obj["debug_probe_params"], pass_obj["debug_mailbox_params"], message)
    click.echo("Set fault analysis mode succeeded.")


def famode(
    debug_probe_params: DebugProbeParams,
    debug_mailbox_params: DebugMailboxParams,
    message: str,
) -> None:
    """Set Fault Analysis Mode.

    :param debug_probe_params: DebugProbeParams object holding information about parameters for debug probe.
    :param debug_mailbox_params: DebugMailboxParams object holding information about parameters for debugmailbox.
    :param message: Path to message file..
    :raises SPSDKAppError: Raised if any error occurred.
    """
    try:
        with _open_debugmbox(debug_probe_params, debug_mailbox_params) as mail_box:
            if message:
                data = load_binary(message)
                data_words = list(struct.unpack(f"<{len(data) // 4}I", data))
                dm_commands.SetFaultAnalysisMode(dm=mail_box, paramlen=len(data_words)).run(
                    data_words
                )
            else:
                dm_commands.SetFaultAnalysisMode(dm=mail_box).run()
    except Exception as e:
        raise SPSDKAppError(f"Set fault analysis mode failed: {e}") from e


@main.command(name="ispmode", no_args_is_help=True)
@click.option("-m", "--mode", type=INT(), required=True)
@click.pass_obj
def ispmode_command(pass_obj: dict, mode: int) -> None:
    """Enter ISP Mode."""
    ispmode(pass_obj["debug_probe_params"], pass_obj["debug_mailbox_params"], mode)
    click.echo("Entering into ISP mode succeeded")


def ispmode(
    debug_probe_params: DebugProbeParams, debug_mailbox_params: DebugMailboxParams, mode: int
) -> None:
    """Enter ISP Mode.

    :param debug_probe_params: DebugProbeParams object holding information about parameters for debug probe.
    :param debug_mailbox_params: DebugMailboxParams object holding information about parameters for debugmailbox.
    :param mode: ISP mode
    :raises SPSDKAppError: Raised if any error occurred.
    """
    try:
        with _open_debugmbox(debug_probe_params, debug_mailbox_params) as mail_box:
            dm_commands.EnterISPMode(dm=mail_box).run([mode])
    except Exception as e:
        raise SPSDKAppError(f"Entering into ISP mode failed: {e}") from e


@main.command(name="token_auth", no_args_is_help=True)
@click.option(
    "-f", "--file", type=click.Path(), required=True, help="Path to token file (string hex format)."
)
@click.option(
    "-n",
    "--no-exit",
    is_flag=True,
    help="When used, exit debug mailbox command is not executed after debug authentication.",
)
@click.pass_obj
def token_auth_command(pass_obj: dict, file: str, no_exit: bool) -> None:
    """Debug Authentication using token."""
    token_auth(pass_obj["debug_probe_params"], pass_obj["debug_mailbox_params"], file, no_exit)
    click.echo("Debug authentication using token succeeded")


def token_auth(
    debug_probe_params: DebugProbeParams,
    debug_mailbox_params: DebugMailboxParams,
    token_file: str,
    no_exit: bool,
) -> None:
    """Debug Authentication using token.

    :param debug_probe_params: DebugProbeParams object holding information about parameters for debug probe.
    :param debug_mailbox_params: DebugMailboxParams object holding information about parameters for debugmailbox.
    :param token_file: Path to token file.
    :param no_exit: When true, exit debug mailbox command is not executed after debug authentication.
    :raises SPSDKAppError: Raised if any error occurred.
    """
    token = []
    logger.info("Starting Debug Authentication for Blank Device..")
    try:
        with _open_debugmbox(debug_probe_params, debug_mailbox_params) as mail_box:
            with open(token_file, "rb") as f:
                while True:
                    chunk = f.read(8).strip()
                    if not chunk:
                        break
                    token.append(int(chunk, 16))

            dm_commands.EnterBlankDebugAuthentication(dm=mail_box).run(token)
            if not no_exit:
                exit_response = dm_commands.ExitDebugMailbox(dm=mail_box).run()
                logger.debug(f"Exit response: {exit_response}")
                # Re-open debug probe
                mail_box.debug_probe.close()
                mail_box.debug_probe.open()
                # Do test of access to AHB bus
                ahb_access_granted = test_ahb_access(mail_box.debug_probe)
                res_str = (
                    (colorama.Fore.GREEN + "successfully")
                    if ahb_access_granted
                    else (colorama.Fore.RED + "without AHB access")
                )
                if not ahb_access_granted:
                    raise SPSDKAppError("Access to AHB is not granted.")

                logger.info(
                    f"Debug Authentication using token ends {res_str}{colorama.Fore.RESET}."
                )
            else:
                logger.info(
                    "Debug Authentication using token ends without exit and without test of AHB access."
                )
    except Exception as e:
        raise SPSDKAppError(f"Debug authentication using token  failed: {e}") from e


@main.command(name="get-crp")
@click.pass_obj
def get_crp_command(pass_obj: dict) -> None:
    """Get CRP level."""
    get_crp(pass_obj["debug_probe_params"], pass_obj["debug_mailbox_params"])
    click.echo("Get CRP Level succeeded")


def get_crp(
    debug_probe_params: DebugProbeParams,
    debug_mailbox_params: DebugMailboxParams,
) -> None:
    """Get CRP level. This command should be called after 'start' command and with no-reset '-n' option.

    :param debug_probe_params: DebugProbeParams object holding information about parameters for debug probe.
    :param debug_mailbox_params: DebugMailboxParams object holding information about parameters for debugmailbox.
    :raises SPSDKAppError: Raised if any error occurred.
    """
    try:
        with _open_debugmbox(debug_probe_params, debug_mailbox_params) as mail_box:
            crp_level = dm_commands.GetCRPLevel(dm=mail_box).run()[0]
            logger.info(f"CRP level is: {crp_level}.")
    except Exception as e:
        raise SPSDKAppError(f"Get CRP Level failed: {e}") from e


@main.command(name="start-debug-session")
@click.pass_obj
def start_debug_session_command(pass_obj: dict) -> None:
    """Start debug session."""
    start_debug_session(pass_obj["debug_probe_params"], pass_obj["debug_mailbox_params"])
    click.echo("Start debug session succeeded")


def start_debug_session(
    debug_probe_params: DebugProbeParams,
    debug_mailbox_params: DebugMailboxParams,
) -> None:
    """Start debug session.

    :param debug_probe_params: DebugProbeParams object holding information about parameters for debug probe.
    :param debug_mailbox_params: DebugMailboxParams object holding information about parameters for debugmailbox.
    :raises SPSDKAppError: Raised if any error occurred.
    """
    try:
        with _open_debugmbox(debug_probe_params, debug_mailbox_params) as mail_box:
            dm_commands.StartDebugSession(dm=mail_box).run()
    except Exception as e:
        raise SPSDKAppError(f"Start debug session failed: {e}") from e


@main.command(name="test-connection")
@click.pass_obj
def test_connection_command(pass_obj: dict) -> None:
    """Method just try if the device debug port is opened or not."""
    ahb_access_granted = test_connection(pass_obj["debug_probe_params"])
    access_str = colorama.Fore.GREEN if ahb_access_granted else colorama.Fore.RED + "not-"
    click.echo(f"The device is {access_str}accessible for debugging.{colorama.Fore.RESET}")


def test_connection(debug_probe_params: DebugProbeParams) -> bool:
    """Method just try if the device debug port is opened or not.

    :param debug_probe_params: DebugProbeParams object holding information about parameters for debug probe.
    :raises SPSDKAppError: Raised if any error occurred.
    """
    try:
        with open_debug_probe(
            interface=debug_probe_params.interface,
            serial_no=debug_probe_params.serial_no,
            debug_probe_params=debug_probe_params.debug_probe_user_params,
        ) as debug_probe:
            ahb_access_granted = test_ahb_access(debug_probe)
        return ahb_access_granted
    except Exception as e:
        raise SPSDKAppError(f"Testing AHB access failed: {e}") from e


@main.command(name="read-memory", no_args_is_help=True)
@click.argument("address", type=INT(), required=True)
@click.argument("byte_count", type=INT(), required=True)
@click.argument("out_file", metavar="FILE", type=click.Path(), required=False)
@click.option("-h", "--use-hexdump", is_flag=True, default=False, help="Use hexdump format")
@click.pass_obj
def read_memory_command(
    pass_obj: dict,
    address: int,
    byte_count: int,
    out_file: str,
    use_hexdump: bool,
) -> None:
    """Reads the memory and writes it to the file or stdout.

    Returns the contents of memory at the given <ADDRESS>, for a specified <BYTE_COUNT>.
    Data are read by 4 bytes at once and are store in little endian format!
    \b
    ADDRESS     - starting address
    BYTE_COUNT  - number of bytes to read
    FILE        - store result into this file, if not specified use stdout
    """
    with progress_bar(suppress=logger.getEffectiveLevel() > logging.INFO) as progress_callback:
        data = read_memory(pass_obj["debug_probe_params"], address, byte_count, progress_callback)
    if out_file:
        write_file(data, out_file, mode="wb")
        click.echo(f"The memory has been read and written into {out_file}")
    else:
        click.echo(format_raw_data(data, use_hexdump=use_hexdump))


def read_memory(
    debug_probe_params: DebugProbeParams,
    address: int,
    byte_count: int,
    progress_callback: Optional[Callable[[int, int], None]] = None,
) -> bytes:
    """Reads the memory.

    :param debug_probe_params: DebugProbeParams object holding information about parameters for debug probe.
    :param address: Starting address.
    :param byte_count: Number of bytes to read.
    :param progress_callback: Progressbar callback method.
    :raises SPSDKAppError: Raised if any error occurred.
    """
    bin_image = BinaryImage("memRead", byte_count, offset=address)
    start_addr = bin_image.aligned_start(4)
    length = bin_image.aligned_length(4)

    data = bytes()
    with open_debug_probe(
        interface=debug_probe_params.interface,
        serial_no=debug_probe_params.serial_no,
        debug_probe_params=debug_probe_params.debug_probe_user_params,
    ) as debug_probe:
        try:
            for addr in range(start_addr, start_addr + length, 4):
                if progress_callback:
                    progress_callback(addr, start_addr + length)
                data += debug_probe.mem_reg_read(addr).to_bytes(4, "little")
        except SPSDKError as exc:
            raise SPSDKAppError(str(exc)) from exc

    if not data:
        raise SPSDKAppError("The read operation failed.")
    if len(data) != length:
        logger.warning(
            f"The memory wasn't read complete. It was read just first {len(data) - (address-start_addr)} Bytes."
        )
    # Shrink start padding data
    data = data[address - start_addr :]
    # Shrink end padding data
    data = data[:byte_count]
    return data


@main.command(name="write-memory", no_args_is_help=True)
@click.argument("address", type=INT(), required=True)
@click.argument("data_source", metavar="FILE[,BYTE_COUNT] | {{HEX-DATA}}", type=str, required=True)
@click.pass_obj
def write_memory_command(pass_obj: dict, address: int, data_source: str) -> None:
    """Writes memory from a file or a hex-data.

    Writes memory at <ADDRESS> from <FILE> or <HEX-DATA>
    Writes a provided buffer to a specified <BYTE_COUNT> in memory.

    \b
    ADDRESS     - starting address
    FILE        - write the content of this file
    BYTE_COUNT  - if specified, load only first BYTE_COUNT number of bytes from file
    HEX-DATA    - string of hex values: {{112233}}, "{{11 22 33}}"
                - when using Jupyter notebook, use [[ ]] instead of {{ }}: eg. [[112233]]
    """
    try:
        data = parse_hex_data(data_source)
    except SPSDKError:
        file_path, size = parse_file_and_size(data_source)
        with open(file_path, "rb") as f:
            data = f.read(size)
    write_memory(pass_obj["debug_probe_params"], address, data)
    click.echo("The memory has been written successfully.")


def write_memory(debug_probe_params: DebugProbeParams, address: int, data: bytes) -> None:
    """Writes memory from a file or a hex-data.

    :param debug_probe_params: DebugProbeParams object holding information about parameters for debug probe.
    :param address: Starting address.
    :param data: Data to write into memory.
    :raises SPSDKAppError: Raised if any error occurred.
    """
    byte_count = len(data)
    bin_image = BinaryImage("memRead", byte_count, offset=address)
    start_addr = bin_image.aligned_start(4)
    length = bin_image.aligned_length(4)
    with open_debug_probe(
        interface=debug_probe_params.interface,
        serial_no=debug_probe_params.serial_no,
        debug_probe_params=debug_probe_params.debug_probe_user_params,
    ) as debug_probe:
        start_padding = address - start_addr
        align_data = data
        if start_padding:
            align_start_word = debug_probe.mem_reg_read(start_addr).to_bytes(4, "little")
            align_data = align_start_word[:start_padding] + data

        end_padding = length - byte_count - start_padding
        if end_padding:
            align_end_word = debug_probe.mem_reg_read(start_addr + length - 4).to_bytes(4, "little")
            align_data = align_data + align_end_word[4 - end_padding :]

        with progress_bar(suppress=logger.getEffectiveLevel() > logging.INFO) as progress_callback:
            for i, addr in enumerate(range(start_addr, start_addr + length, 4)):
                progress_callback(addr, start_addr + length)
                to_write = int.from_bytes(align_data[i * 4 : i * 4 + 4], "little")
                debug_probe.mem_reg_write(addr, to_write)
                # verify write
                try:
                    verify_data = debug_probe.mem_reg_read(addr)
                except SPSDKError as ver_exc:
                    raise SPSDKAppError("The write verification failed.") from ver_exc
                if to_write != verify_data:
                    raise SPSDKAppError(
                        f"Data verification failed! {hex(to_write)} != {hex(verify_data)}"
                    )


@main.command(name="get-uuid")
@click.pass_obj
def get_uuid_command(pass_obj: dict) -> None:
    """Get the UUID from target if possible.

    Some devices need to call 'start' command prior the get-uuid!
    Also there could be issue with repeating of this command without hard reset of device 'reset -h'.
    """
    uuid = get_uuid(pass_obj["debug_probe_params"], pass_obj["debug_mailbox_params"])
    if uuid:
        click.echo(f"The device UUID is: {uuid.hex()}")
    else:
        click.echo("The device UUID is not possible to retrieve from target.")


def get_uuid(
    debug_probe_params: DebugProbeParams, debug_mailbox_params: DebugMailboxParams
) -> Optional[bytes]:
    """Get the UUID from target if possible.

    Some devices need to call 'start' command prior the get-uuid!
    Also there could be issue with repeating of this command without hard reset of device 'reset -h'.

    :param debug_probe_params: DebugProbeParams object holding information about parameters for debug probe.
    :param debug_mailbox_params: DebugMailboxParams object holding information about parameters for debug mailbox.
    :raises SPSDKAppError: Raised if any error occurred.
    :return: UUID value in bytes if succeeded, None otherwise.
    """
    try:
        with open_debug_probe(
            debug_probe_params.interface,
            debug_probe_params.serial_no,
            debug_probe_params.debug_probe_user_params,
        ) as debug_probe:
            try:
                dm = DebugMailbox(
                    debug_probe=debug_probe,
                    reset=debug_mailbox_params.reset,
                    moredelay=debug_mailbox_params.more_delay,
                    op_timeout=debug_mailbox_params.operation_timeout,
                )
                dac_data = dm_commands.DebugAuthenticationStart(dm=dm, resplen=26).run()
            except SPSDKError:
                debug_probe.close()
                debug_probe.open()
                dm = DebugMailbox(
                    debug_probe=debug_probe,
                    reset=debug_mailbox_params.reset,
                    moredelay=debug_mailbox_params.more_delay,
                    op_timeout=debug_mailbox_params.operation_timeout,
                )
                dac_data = dm_commands.DebugAuthenticationStart(dm=dm, resplen=30).run()
            # convert List[int] to bytes
            dac_data_bytes = struct.pack(f"<{len(dac_data)}I", *dac_data)
            dac = DebugAuthenticationChallenge.parse(dac_data_bytes)
    except Exception as e:
        raise SPSDKAppError(f"Getting UUID from target failed: {e}") from e

    if dac.uuid == bytes(16):
        logger.warning("The valid UUID is not included in DAC.")
        logger.info(f"DAC info:\n {dac.info()}")
        return None

    logger.info(
        f"Got DAC from SOCC:'{DebugCredential.get_socc_description(dac.version, dac.socc)}' to retrieve UUID."
    )
    return dac.uuid


@main.command(name="gendc", no_args_is_help=True)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, dir_okay=False),
    required=True,
    help="Specify YAML credential config file.",
)
@click.option(
    "-e",
    "--elf2sb-config",
    type=click.Path(exists=True, dir_okay=False),
    required=False,
    help="Specify Root Of Trust from configuration file used by elf2sb tool",
)
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="Force overwriting of an existing file. Create destination folder, if doesn't exist already.",
)
@spsdk_plugin_option
@click.argument("dc_file_path", metavar="PATH", type=click.Path(file_okay=True))
@click.pass_obj
def gendc_command(
    pass_obj: dict,
    plugin: str,
    dc_file_path: str,
    config: str,
    elf2sb_config: str,
    force: bool,
) -> None:
    """Generate debug certificate (DC).

    The -p protocol option must be defined in main application.

    \b
    PATH    - path to dc file
    """
    gendc(
        pass_obj["protocol"],
        plugin,
        dc_file_path,
        config,
        elf2sb_config,
        force,
    )
    click.echo("Creating Debug credential file succeeded")


def gendc(
    protocol: DatProtocol,
    plugin: str,
    dc_file_path: str,
    config: str,
    elf2sb_config: str,
    force: bool,
) -> None:
    """Generate debug certificate (DC).

    :param protocol: Debug authentication protocol.
    :param plugin: Path to external python file containing a custom SignatureProvider implementation.
    :param dc_file_path: Path to debug certificate file.
    :param config: YAML credential config file.
    :param elf2sb_config: Root Of Trust from configuration file used by elf2sb tool.
    :param force: Force overwriting of an existing file.
    :raises SPSDKAppError: Raised if any error occurred.
    """
    protocol.validate()
    try:
        if plugin:
            load_plugin_from_source(plugin)
        check_file_exists(dc_file_path, force)
        logger.info("Loading configuration from yml file...")
        yaml_content = load_configuration(config)
        if elf2sb_config:
            elf2sb_config_dir = os.path.dirname(elf2sb_config)
            logger.info("Loading configuration from elf2sb config file...")
            rot_info = RootOfTrustInfo(
                load_configuration(elf2sb_config), search_paths=[elf2sb_config_dir]
            )
            yaml_content["rot_meta"] = [
                find_file(x, search_paths=[elf2sb_config_dir]) for x in rot_info.public_keys
            ]
            assert rot_info.private_key or rot_info.signature_provider_cfg
            if rot_info.private_key:
                yaml_content["rotk"] = find_file(
                    rot_info.private_key, search_paths=[elf2sb_config_dir]
                )
            if rot_info.signature_provider_cfg:
                yaml_content["sign_provider"] = rot_info.signature_provider_cfg
            yaml_content["rot_id"] = rot_info.public_key_index

        # enforcing rot_id presence in yaml config...
        assert "rot_id" in yaml_content, "Config file doesn't contain the 'rot_id' field"

        logger.info(f"Creating {'RSA' if protocol.is_rsa() else 'ECC'} debug credential object...")
        dc = DebugCredential.create_from_yaml_config(
            version=protocol.version,
            yaml_config=yaml_content,
            search_paths=[os.path.dirname(config)],
        )
        dc.sign()
        data = dc.export()
        click.echo(f"RoT Key Hash: {dc.get_rotkh().hex()}")
        logger.debug(f"Debug credential file details:\n {dc.info()}")
        logger.info(f"Saving the debug credential to a file: {dc_file_path}")
        write_file(data, dc_file_path, mode="wb")

    except Exception as e:
        raise SPSDKAppError(f"The generating of Debug Credential file failed: {e}") from e


@main.command(name="get-template", no_args_is_help=True)
@click.argument("output", metavar="PATH", type=click.Path())
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force overwriting of an existing file. Create destination folder, if doesn't exist already.",
)
def get_template_command(output: str, force: bool) -> None:
    """Generate the template of Debug Credentials YML configuration file.

    \b
    PATH    - file name path to write template config file
    """
    get_template(output, force)
    click.echo("The configuration template file has been created.")


def get_template(output: str, force: bool) -> None:
    """Generate the template of Debug Credentials YML configuration file.

    :param output: Path to output file.
    :param force: Force overwriting of an existing file.
    """
    check_file_exists(str(output), force)
    write_file(load_text(os.path.join(NXPDEBUGMBOX_DATA_FOLDER, "template_config.yaml")), output)


@catch_spsdk_error
def safe_main() -> None:
    """Call the main function."""
    sys.exit(main())  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    safe_main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module for DebugMailbox PyLink Debug probes support."""

import logging
from typing import Dict, Optional

import pylink
from pylink.errors import JLinkException

from spsdk.utils.misc import value_to_int

from .debug_probe import (
    DebugProbe,
    SPSDKDebugProbeError,
    SPSDKDebugProbeNotOpenError,
    SPSDKDebugProbeTransferError,
)

logger = logging.getLogger(__name__)
JLINK_LOGGER = logger.getChild("PyLink")
JLINK_LOGGER.setLevel(logging.ERROR)


def set_logger(level: int) -> None:
    """Sets the log level for this module.

    param level: Requested level.
    """
    logger.setLevel(level)


set_logger(logging.ERROR)


class DebugProbePyLink(DebugProbe):
    """Class to define PyLink package interface for NXP SPSDK."""

    @staticmethod
    def get_options_help() -> Dict[str, str]:
        """Get full list of options of debug probe.

        :return: Dictionary with individual options. Key is parameter name and value the help text.
        """
        return {
            "frequency": "Set the communication frequency in KHz, default is 100KHz",
        }

    @classmethod
    def get_jlink_lib(cls) -> pylink.JLink:
        """Get J-Link object.

        :return: The J-Link Object
        :raises SPSDKDebugProbeError: The J-Link object get function failed.
        """
        try:
            return pylink.JLink(
                log=JLINK_LOGGER.info,
                detailed_log=JLINK_LOGGER.debug,
                error=JLINK_LOGGER.error,
                warn=JLINK_LOGGER.warn,
            )
        except TypeError as exc:
            raise SPSDKDebugProbeError("Cannot open Jlink DLL") from exc

    def __init__(self, hardware_id: str, options: Optional[Dict] = None) -> None:
        """The PyLink class initialization.

        The PyLink initialization function for SPSDK library to support various DEBUG PROBES.
        """
        super().__init__(hardware_id, options)
        set_logger(logging.root.level)
        self.pylink = None

        logger.debug("The SPSDK PyLink Interface has been initialized")

    @classmethod
    def get_connected_probes(
        cls, hardware_id: Optional[str] = None, options: Optional[Dict] = None
    ) -> list:
        """Get all connected probes over PyLink.

        This functions returns the list of all connected probes in system by PyLink package.

        :param hardware_id: None to list all probes, otherwise the the only probe with matching
            hardware id is listed.
        :param options: The options dictionary
        :return: probe_description
        """
        # pylint: disable=import-outside-toplevel
        from .utils import DebugProbes, ProbeDescription

        jlink = DebugProbePyLink.get_jlink_lib()

        probes = DebugProbes()
        connected_probes = jlink.connected_emulators()
        for probe in connected_probes:
            if not hardware_id or hardware_id == str(probe.SerialNumber):
                probes.append(
                    ProbeDescription(
                        "Jlink",
                        str(probe.SerialNumber),
                        f"Segger {probe.acProduct.decode('utf-8')}: {str(probe.SerialNumber)}",
                        DebugProbePyLink,
                    )
                )

        return probes

    def open(self) -> None:
        """Open PyLink interface for NXP SPSDK.

        The PyLink opening function for SPSDK library to support various DEBUG PROBES.
        The function is used to initialize the connection to target and enable using debug probe
        for DAT purposes.

        :raises SPSDKDebugProbeError: The PyLink cannot establish communication with target
        """
        try:
            self.pylink = DebugProbePyLink.get_jlink_lib()
            if self.pylink is None:
                raise SPSDKDebugProbeError("Getting of J-Link library failed")
        except SPSDKDebugProbeError as exc:
            raise SPSDKDebugProbeError(f"Getting of J-Link library failed({str(exc)})") from exc

        try:
            self.pylink.open(
                serial_no=self.hardware_id,
                ip_addr=self.options.get("ip_address") if self.options else None,
            )
            if self.options.get("use_jtag") is None:
                self.pylink.set_tif(pylink.enums.JLinkInterfaces.SWD)
            else:
                logger.warning(
                    "Experimental support for JTAG on RW61x."
                    "The implementation may have bugs and lack features."
                )
                self.pylink.set_speed(100)
                self.pylink.set_tif(pylink.enums.JLinkInterfaces.JTAG)
            self.pylink.coresight_configure()
            self.pylink.set_speed(speed=value_to_int(self.options.get("frequency", 100)))
            # Power Up the system and debug and clear sticky errors
            if self.options.get("use_jtag") is not None:
                # Currently clear_sticky_errors has been defined only for SWD (uncleared for JTAG-DP)
                self.coresight_reg_write(access_port=False, addr=4, data=0x50000F20)
            self.clear_sticky_errors()
            self.power_up_target()

        except JLinkException as exc:
            raise SPSDKDebugProbeError(
                f"PyLink cannot establish communication with target({str(exc)})"
            ) from exc

    def close(self) -> None:
        """Close PyLink interface.

        The PyLink closing function for SPSDK library to support various DEBUG PROBES.
        """
        if self.pylink:
            self.pylink.close()

    def coresight_reg_read(self, access_port: bool = True, addr: int = 0) -> int:
        """Read coresight register over PyLink interface.

        The PyLink read coresight register function for SPSDK library to support various DEBUG PROBES.

        :param access_port: if True, the Access Port (AP) register will be read(default), otherwise the Debug Port
        :param addr: the register address
        :return: The read value of addressed register (4 bytes)
        :raises SPSDKDebugProbeTransferError: The IO operation failed
        :raises SPSDKDebugProbeNotOpenError: The PyLink probe is NOT opened

        """
        if self.pylink is None:
            raise SPSDKDebugProbeNotOpenError("The PyLink debug probe is not opened yet")

        try:
            if access_port:
                self.select_ap(addr)
                addr = addr & 0x0F

            return self.pylink.coresight_read(reg=addr // 4, ap=access_port)
        except (JLinkException, ValueError, TypeError) as exc:
            # In case of transaction error reconfigure and initialize the JLink
            self._reinit_jlink_target()
            raise SPSDKDebugProbeTransferError(
                f"The Coresight read operation failed({str(exc)})"
            ) from exc

    def coresight_reg_write(self, access_port: bool = True, addr: int = 0, data: int = 0) -> None:
        """Write coresight register over PyLink interface.

        The PyLink write coresight register function for SPSDK library to support various DEBUG PROBES.

        :param access_port: if True, the Access Port (AP) register will be write(default), otherwise the Debug Port
        :param addr: the register address
        :param data: the data to be written into register
        :raises SPSDKDebugProbeTransferError: The IO operation failed
        :raises SPSDKDebugProbeNotOpenError: The PyLink probe is NOT opened
        """
        if self.pylink is None:
            raise SPSDKDebugProbeNotOpenError("The PyLink debug probe is not opened yet")

        try:
            if access_port:
                self.select_ap(addr)
                addr = addr & 0x0F

            self.pylink.coresight_write(reg=addr // 4, data=data, ap=access_port)

        except (JLinkException, ValueError, TypeError) as exc:
            # In case of transaction error reconfigure and initialize the JLink
            self._reinit_jlink_target()
            raise SPSDKDebugProbeTransferError(
                f"The Coresight write operation failed({str(exc)})"
            ) from exc

    def assert_reset_line(self, assert_reset: bool = False) -> None:
        """Control reset line at a target.

        :param assert_reset: If True, the reset line is asserted(pulled down), if False the reset line is not affected.
        :raises SPSDKDebugProbeNotOpenError: The PyLink probe is NOT opened
        :raises SPSDKDebugProbeError: The PyLink probe RESET function failed
        """
        if self.pylink is None:
            raise SPSDKDebugProbeNotOpenError("The PyLink debug probe is not opened yet")

        try:
            if assert_reset:
                self.pylink.set_reset_pin_low()
            else:
                self.pylink.set_reset_pin_high()
        except JLinkException as exc:
            raise SPSDKDebugProbeError(f"Jlink reset operation failed: {str(exc)}") from exc

    def _reinit_jlink_target(self) -> None:
        """Re-initialize the Jlink connection."""
        if not self.disable_reinit:
            assert self.pylink
            self.pylink.coresight_configure()
            if self.options.get("use_jtag") is not None:
                # Currently clear_sticky_errors has been defined only for SWD (uncleared for JTAG-DP)
                self.coresight_reg_write(access_port=False, addr=4, data=0x50000F20)
            self._reinit_target()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module for certificate management (generating certificate, validating certificate, chains)."""

from datetime import datetime, timedelta
from typing import Dict, List, Optional, Union

from spsdk import SPSDKError
from spsdk.crypto import (
    Certificate,
    CertificateSigningRequest,
    Encoding,
    ExtensionOID,
    InvalidSignature,
    PrivateKey,
    PublicKey,
    _PublicKeyTuple,
    default_backend,
    hashes,
    padding,
    rsa,
    x509,
)
from spsdk.utils.misc import write_file


def generate_certificate(
    subject: x509.Name,
    issuer: x509.Name,
    subject_public_key: PublicKey,
    issuer_private_key: PrivateKey,
    serial_number: Optional[int] = None,
    if_ca: bool = True,
    duration: int = 3650,
    path_length: int = 2,
) -> Certificate:
    """Generate certificate.

    :param subject: subject name that the CA issues the certificate to
    :param issuer: issuer name that issued the certificate
    :param subject_public_key: RSA public key of subject
    :param issuer_private_key: RSA private key of issuer
    :param serial_number: certificate serial number, if not specified, random serial number will be set
    :param if_ca: true if the certificate can sign certificates, none otherwise
    :param duration: how long the certificate will be valid (in days)
    :param path_length: The maximum path length for certificates subordinate to this certificate.
    :return: certificate
    """
    crt = x509.CertificateBuilder(
        subject_name=subject,
        issuer_name=issuer,
        not_valid_before=datetime.utcnow(),
        not_valid_after=datetime.utcnow() + timedelta(days=duration),
        public_key=subject_public_key,
        extensions=[],
        serial_number=serial_number or x509.random_serial_number(),
    )

    crt = crt.add_extension(
        x509.BasicConstraints(ca=if_ca, path_length=path_length if if_ca else None),
        critical=True,
    )
    return crt.sign(issuer_private_key, hashes.SHA256(), default_backend())


def save_crypto_item(
    item: Union[Certificate, CertificateSigningRequest],
    file_path: str,
    encoding_type: Encoding = Encoding.PEM,
) -> None:
    """Save the certificate/CSR into file.

    :param item: certificate or certificate signing request
    :param file_path: path to the file where item will be stored
    :param encoding_type: encoding type (PEM or DER)
    """
    write_file(item.public_bytes(encoding_type), file_path, mode="wb")


def validate_certificate_chain(chain_list: list) -> list:
    """Validate chain of certificates.

    :param chain_list: list of certificates in chain
    :return: list of boolean values, which corresponds to the certificate validation in chain
    :raises SPSDKError: When chain has less than two certificates
    """
    if len(chain_list) <= 1:
        raise SPSDKError("The chain must have at least two certificates")
    result = []
    for i in range(len(chain_list) - 1):
        result.append(validate_certificate(chain_list[i], chain_list[i + 1]))
    return result


def validate_certificate(subject_certificate: Certificate, issuer_certificate: Certificate) -> bool:
    """Validate certificate.

    :param subject_certificate: subject's certificate
    :param issuer_certificate: issuer's certificate
    :raises SPSDKError: Unsupported key type in Certificate
    :return: true/false whether certificate is valid or not
    """
    issuer_pub_key = get_public_key_from_certificate(issuer_certificate)
    cert_to_check = x509.load_pem_x509_certificate(
        convert_certificate_into_bytes(subject_certificate), default_backend()
    )

    try:
        if isinstance(issuer_pub_key, rsa.RSAPublicKey):
            assert cert_to_check.signature_hash_algorithm
            issuer_pub_key.verify(
                cert_to_check.signature,
                cert_to_check.tbs_certificate_bytes,
                padding.PKCS1v15(),
                cert_to_check.signature_hash_algorithm,
            )
            return True
        raise SPSDKError(f"Certificate validation for {type(issuer_pub_key)} is not supported")
    except InvalidSignature:
        return False


def is_ca_flag_set(certificate: Certificate) -> bool:
    """Check if CA flag is set in certificate.

    :param certificate: Certificate to be checked
    :return: true/false depending whether ca flag is set or not
    """
    extension = certificate.extensions.get_extension_for_oid(ExtensionOID.BASIC_CONSTRAINTS)
    return extension.value.ca  # type: ignore # mypy can not handle property definition in cryptography


def validate_ca_flag_in_cert_chain(chain_list: List[Certificate]) -> bool:
    """Validate CA flag in certification chain.

    :param chain_list: list of certificates in the chain
    :return: true/false depending whether ca flag is set or not
    """
    return is_ca_flag_set(chain_list[0])


def get_public_key_from_certificate(certificate: Certificate) -> PublicKey:
    """Get public keys from certificate.

    :param certificate: certificate item
    :return: RSA public key
    """
    public_key = certificate.public_key()
    assert isinstance(public_key, _PublicKeyTuple)
    return public_key


def convert_certificate_into_bytes(
    certificate: Certificate, encoding: Encoding = Encoding.PEM
) -> bytes:
    """Convert certificates into bytes.

    :param certificate: certificate item
    :param encoding: encoding type
    :return: certificate in bytes form
    """
    assert isinstance(certificate, Certificate), "The input is not a Certificate"
    return certificate.public_bytes(encoding)


X509NameConfig = Union[List[Dict[str, str]], Dict[str, Union[str, List[str]]]]


def generate_name(config: X509NameConfig) -> x509.Name:
    """Generate x509 Name.

    :param config: subject/issuer description
    :return: x509.Name
    """
    attributes: List[x509.NameAttribute] = []

    def _get_name_oid(name: str) -> x509.ObjectIdentifier:
        try:
            return getattr(x509.NameOID, name)
        except Exception as exc:
            raise SPSDKError(f"Invalid value of certificate attribute: {name}") from exc

    if isinstance(config, list):
        for item in config:
            for key, value in item.items():
                name_oid = _get_name_oid(key)
                attributes.append(x509.NameAttribute(name_oid, str(value)))

    if isinstance(config, dict):
        for key_second, value_second in config.items():
            name_oid = _get_name_oid(key_second)
            if isinstance(value_second, list):
                for value in value_second:
                    attributes.append(x509.NameAttribute(name_oid, str(value)))
            else:
                attributes.append(x509.NameAttribute(name_oid, str(value_second)))

    return x509.Name(attributes)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module is used to generate initialization SB file."""

import logging
import os
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple

import click

from spsdk import SPSDK_DATA_FOLDER, SPSDKError, SPSDKValueError
from spsdk.apps.utils import spsdk_logger
from spsdk.apps.utils.common_cli_options import (
    CommandsTreeGroup,
    isp_interfaces,
    spsdk_apps_common_options,
)
from spsdk.apps.utils.utils import catch_spsdk_error, format_raw_data, get_interface
from spsdk.image import MBIMG_SCH_FILE
from spsdk.mboot.commands import TrustProvKeyType, TrustProvOemKeyType
from spsdk.mboot.interfaces import MBootInterface
from spsdk.mboot.mcuboot import McuBoot
from spsdk.sbfile.sb31.commands import CmdLoadKeyBlob
from spsdk.sbfile.sb31.images import (
    SB3_SCH_FILE,
    SecureBinary31,
    SecureBinary31Commands,
    SecureBinary31Header,
)
from spsdk.utils.crypto.cert_blocks import CertificateBlockHeader
from spsdk.utils.crypto.common import crypto_backend
from spsdk.utils.database import Database
from spsdk.utils.misc import (
    load_binary,
    load_configuration,
    value_to_bool,
    value_to_int,
    write_file,
)
from spsdk.utils.schema_validator import ConfigTemplate, ValidationSchemas, check_config

logger = logging.getLogger(__name__)

NXPDEVHSM_DATA_FOLDER: str = os.path.join(SPSDK_DATA_FOLDER, "nxpdevhsm")
NXPDEVHSM_DATABASE_FILE: str = os.path.join(NXPDEVHSM_DATA_FOLDER, "database.yaml")


class DeviceHsm:
    """Class to handle device HSM provisioning procedure."""

    DEVBUFF_SIZE = 0x100
    DEVBUFF_GEN_MASTER_SHARE_INPUT_SIZE = 16
    DEVBUFF_GEN_MASTER_ENC_SHARE_OUTPUT_SIZE = 48
    DEVBUFF_GEN_MASTER_ENC_MASTER_SHARE_OUTPUT_SIZE = 64
    DEVBUFF_GEN_MASTER_CUST_CERT_PUK_OUTPUT_SIZE = 64
    DEVBUFF_HSM_GENKEY_KEYBLOB_SIZE = 48
    DEVBUFF_HSM_GENKEY_KEYBLOB_PUK_SIZE = 64
    DEVBUFF_CUST_MK_SK_KEY_SIZE = 32
    DEVBUFF_WRAPPED_CUST_MK_SK_KEY_SIZE = 48
    DEVBUFF_DATA_BLOCK_SIZE = 256
    DEVBUFF_SB3_SIGNATURE_SIZE = 64

    RESET_TIMEOUT = 500  # timeout for reset in milliseconds

    def __init__(
        self,
        mboot: McuBoot,
        cust_mk_sk: bytes,
        oem_share_input: bytes,
        info_print: Callable,
        family: str,
        container_conf: Optional[str] = None,
        workspace: Optional[str] = None,
        initial_reset: bool = False,
        final_reset: bool = True,
    ) -> None:
        """Initialization of device HSM class. Its design to create provisioned SB3 file.

        :param mboot: mBoot communication interface.
        :param oem_share_input: OEM share input data.
        :param cust_mk_sk: Customer Master Key Symmetric Key.
        :param family: chip family
        :param container_conf: Optional elftosb configuration file (to specify user list of SB commands).
        :param workspace: Optional folder to store middle results.
        :param initial_reset: Reset device before DevHSM creation of SB3 file.
        :param final_reset: Reset device after DevHSM creation of SB3 file.
        :raises SPSDKError: In case of any vulnerability.
        """
        self.database = Database(NXPDEVHSM_DATABASE_FILE)
        self.mboot = mboot
        self.cust_mk_sk = cust_mk_sk
        self.oem_share_input = oem_share_input
        self.info_print = info_print
        self.workspace = workspace
        self.initial_reset = initial_reset
        self.final_reset = final_reset
        self.container_conf_dir = os.path.dirname(container_conf) if container_conf else None
        if self.workspace and not os.path.isdir(self.workspace):
            os.mkdir(self.workspace)

        # store input of OEM_SHARE_INPUT to workspace in case that is generated randomly
        self.store_temp_res("OEM_SHARE_INPUT.BIN", self.oem_share_input)

        # Default value that could be given from SB3 configuration container
        self.timestamp = None
        self.sb3_descr = "SB3 SB_KEK"
        self.sb3_fw_ver = 0

        # Check the configuration file and options to update by user config
        self.config_data = None
        if container_conf:
            config_data = load_configuration(container_conf)
            # validate input configuration
            check_config(
                config_data,
                DeviceHsm.get_validation_schemas(family, include_test_configuration=True),
                search_paths=[os.path.dirname(container_conf)],
            )
            self.config_data = config_data
            self.sb3_fw_ver = config_data.get("firmwareVersion") or self.sb3_fw_ver
            self.sb3_descr = config_data.get("description") or self.sb3_descr
            if "timestamp" in config_data:
                self.timestamp = value_to_int(str(config_data.get("timestamp")))

        self.wrapped_cust_mk_sk = bytes()
        self.final_sb = bytes()
        self.family = family
        self.devbuff_base = value_to_int(
            self.database.get_device_value("devbuff_base_address", self.family)
        )

    def get_devbuff_base_address(self, index: int) -> int:
        """Get devbuff base address."""
        assert index < 4 and index >= 0
        return self.devbuff_base + index * self.DEVBUFF_SIZE

    def update_keyblob_offset(self) -> int:
        """Update keyblob offset based on family."""
        return value_to_int(self.database.get_device_value("key_blob_offset", self.family))

    def command_order(self) -> bool:
        """Update command order based on family."""
        return value_to_bool(self.database.get_device_value("order", self.family))

    @staticmethod
    def get_supported_families() -> List[str]:
        """Get the list of supported families by Device HSM.

        :return: List of supported families.
        """
        return Database(NXPDEVHSM_DATABASE_FILE).devices.device_names

    @classmethod
    def get_validation_schemas(
        cls, family: str, include_test_configuration: bool = False
    ) -> List[Dict[str, Any]]:
        """Create the list of validation schemas.

        :param family: Family description.
        :param include_test_configuration: Add also testing configuration schemas.
        :return: List of validation schemas.
        """
        mbi_sch_cfg = ValidationSchemas().get_schema_file(MBIMG_SCH_FILE)
        sb3_sch_cfg = ValidationSchemas().get_schema_file(SB3_SCH_FILE)

        schemas: List[Dict[str, Any]] = []
        schemas.append(mbi_sch_cfg["firmware_version"])
        schemas.append(sb3_sch_cfg["sb3_family"])
        schemas.append(sb3_sch_cfg["sb3_description"])
        schemas.extend(SecureBinary31.get_commands_validation_schemas(family))
        if include_test_configuration:
            schemas.append(sb3_sch_cfg["sb3_test"])
        # find family
        for schema in schemas:
            if "properties" in schema and "family" in schema["properties"]:
                schema["properties"]["family"]["enum"] = cls.get_supported_families()
                break
        return schemas

    @classmethod
    def generate_config_template(cls, family: str) -> Dict[str, str]:
        """Generate configuration for selected family.

        :param family: Family description.
        :return: Dictionary of individual templates (key is name of template, value is template itself).
        """
        ret: Dict[str, str] = {}

        if family in cls.get_supported_families():
            schemas = cls.get_validation_schemas(family)
            schemas.append(ValidationSchemas.get_schema_file(SB3_SCH_FILE)["sb3_output"])
            override = {}
            override["family"] = family

            yaml_data = ConfigTemplate(
                f"DEVHSM procedure Secure Binary v3.1 Configuration template for {family}.",
                schemas,
                override,
            ).export_to_yaml()

            ret[f"sb31_{family}_devhsm"] = yaml_data

        return ret

    def create_sb3(self) -> None:
        """Do device hsm process to create SB_KEK provisioning SB file."""
        # 1: Initial target reset to ensure OEM_MASTER_SHARE works properly (not tainted by previous run)
        if self.initial_reset:
            self.info_print(" 1: Resetting the target device")
            self.mboot.reset(timeout=self.RESET_TIMEOUT)
        else:
            self.info_print(" 1: Initial target reset is disabled ")

        # 2: Call GEN_OEM_MASTER_SHARE to generate encOemShare.bin (ENC_OEM_SHARE will be later put in place of ISK)
        self.info_print(" 2: Generating OEM master share.")
        oem_enc_share, _, _ = self.oem_generate_master_share(self.oem_share_input)

        # 3: Call hsm_gen_key to generate 48 bytes FW signing key
        self.info_print(" 3: Generating 48 bytes FW signing keys.")
        cust_fw_auth_prk, cust_fw_auth_puk = self.generate_key(
            TrustProvOemKeyType.MFWISK, "CUST_FW_AUTH"
        )

        # 4: Call hsm_gen_key to generate 48 bytes FW encryption key
        self.info_print(" 4: Generating 48 bytes FW encryption keys.")
        cust_fw_enc_prk, _ = self.generate_key(TrustProvOemKeyType.MFWENCK, "CUST_FW_ENC_SK")

        # 5: Call hsm_store_key to generate user defined CUST_MK_SK.
        # Will be stored into PFR using loadKeyBlob SB3 command.
        # Use NXP_CUST_KEK_EXT_SK in SB json
        self.info_print(" 5: Wrapping CUST_MK_SK key.")
        self.wrapped_cust_mk_sk = self.wrap_key(self.cust_mk_sk)

        # 6: Generate template sb3 fw, sb3ImageType=6
        self.info_print(" 6: Creating template un-encrypted SB3 header and data blobs.")
        # 6.1: Generate SB3.1 header template
        self.info_print(" 6.1: Creating template SB3 header.")
        sb3_header = SecureBinary31Header(
            firmware_version=self.sb3_fw_ver,
            curve_name="secp256r1",
            description=self.sb3_descr,
            timestamp=self.timestamp,
            flags=0x01,  # Bit0: PROV_MFW: when set, the SB3 file encrypts provisioning firmware
        )
        self.timestamp = sb3_header.timestamp
        sb3_header_exported = sb3_header.export()
        logger.debug(
            f" 6.1: The template SB3 header: \n{sb3_header.info()} \n Length:{len(sb3_header_exported)}"
        )

        # 6.2: Create SB3 file un-encrypted data part
        self.info_print(" 6.2: Creating un-encrypted SB3 data.")
        sb3_data = SecureBinary31Commands(
            family=self.family, curve_name="secp256r1", is_encrypted=False, timestamp=self.timestamp
        )

        if self.container_conf_dir is not None:
            sb3_data.load_from_config(
                self.get_cmd_from_config(), search_paths=[self.container_conf_dir]
            )
        key_blob_command_position = value_to_int(
            self.database.get_device_value("key_blob_command_position", self.family)
        )
        sb3_data.insert_command(
            index=key_blob_command_position,
            command=CmdLoadKeyBlob(
                offset=self.update_keyblob_offset(),
                data=self.wrapped_cust_mk_sk,
                key_wrap_id=CmdLoadKeyBlob.get_key_id(
                    family=self.family, key_name=CmdLoadKeyBlob.KeyTypes.NXP_CUST_KEK_EXT_SK
                ),
            ),
        )

        logger.debug(f" 6.2: Created un-encrypted SB3 data: \n{sb3_data.info()}")
        # 6.3: Get SB3 file data part individual chunks
        data_cmd_blocks = sb3_data.get_cmd_blocks_to_export()

        # 7: Call hsm_enc_blk to encrypt all the data chunks from step 6. Use FW encryption key from step 3.
        self.info_print(" 7: Encrypting SB3 data on device")
        sb3_enc_data = self.encrypt_data_blocks(
            cust_fw_enc_prk, sb3_header_exported, data_cmd_blocks
        )
        # 7.1: Add to encrypted data parts SHA256 hashes
        self.info_print(" 7.1: Enriching encrypted SB3 data by mandatory hashes.")
        enc_final_data = sb3_data.process_cmd_blocks_to_export(sb3_enc_data)
        self.store_temp_res("Final_data.bin", enc_final_data, "to_merge")

        # 7.2: Create dummy certification part of SB3 manifest
        self.info_print(" 7.2: Creating dummy certificate block.")
        cb_header = CertificateBlockHeader()
        cb_header.cert_block_size = (
            cb_header.SIZE + 68 + self.DEVBUFF_GEN_MASTER_ENC_SHARE_OUTPUT_SIZE
        )
        logger.debug(f" 7.2: The dummy certificate block has been created:\n{cb_header.info()}.")

        # 7.3: Update the SB3 pre-prepared header by current data
        self.info_print(" 7.3: Updating SB3 header by valid values.")
        sb3_header.block_count = sb3_data.block_count
        sb3_header.image_total_length += (
            len(sb3_data.final_hash) + cb_header.cert_block_size + self.DEVBUFF_SB3_SIGNATURE_SIZE
        )
        logger.debug(
            f" 7.3: The SB3 header has been updated by valid values:\n{sb3_header.info()}."
        )

        # 7.4: Compose manifest that will be signed
        self.info_print(" 7.4: Preparing SB3 manifest to sign.")
        manifest_to_sign = bytes()
        manifest_to_sign += sb3_header.export()
        manifest_to_sign += sb3_data.final_hash
        manifest_to_sign += cb_header.export()
        manifest_to_sign += (
            b"\x11\x00\x00\x80"  # 0x80000011  Cert Flags: CA Flag, 1 certificate and NIST P-256
        )
        manifest_to_sign += cust_fw_auth_puk
        manifest_to_sign += oem_enc_share
        self.store_temp_res("manifest_to_sign.bin", manifest_to_sign, "to_merge")
        logger.debug(
            f" 7.4: The SB3 manifest data to sign:\n{format_raw_data(manifest_to_sign, use_hexdump=True)}."
        )

        # 8: Get sign of SB3 file manifest
        self.info_print(" 8: Creating SB3 manifest signature on device.")
        manifest_signature = self.sign_data_blob(manifest_to_sign, cust_fw_auth_prk)
        logger.debug(
            f" 8: The SB3 manifest signature data:\n{format_raw_data(manifest_signature, use_hexdump=True)}."
        )

        # 9: Merge all parts together
        self.info_print(" 9: Composing final SB3 file.")
        self.final_sb = bytes()
        self.final_sb += manifest_to_sign
        self.final_sb += manifest_signature
        self.final_sb += enc_final_data
        self.store_temp_res("Final_SB3.sb3", self.final_sb)
        logger.debug(
            f" 9: The final SB3 file data:\n{format_raw_data(self.final_sb, use_hexdump=True)}."
        )

        # 10: Final reset to ensure followup operations (e.g. receive-sb-file) work correctly
        if self.final_reset:
            self.info_print("10: Resetting the target device")
            self.mboot.reset(timeout=self.RESET_TIMEOUT, reopen=False)
        else:
            self.info_print("10: Final target reset disabled")

    def export(self) -> bytes:
        """Get the Final SB file.

        :return: Final SB file in bytes.
        """
        return self.final_sb

    def oem_generate_master_share(self, oem_share_input: bytes) -> Tuple[bytes, bytes, bytes]:
        """Generate on device Encrypted OEM master share outputs.

        :param oem_share_input: OEM input (randomize seed)
        :raises SPSDKError: In case of any vulnerability.
        :return: Tuple with OEM generate master share outputs.
        """
        if not self.mboot.write_memory(self.devbuff_base, oem_share_input):
            raise SPSDKError(
                f"Cannot write OEM SHARE INPUT into device. Error: {self.mboot.status_string}"
            )

        oem_gen_master_share_res = self.mboot.tp_oem_gen_master_share(
            self.devbuff_base,
            self.DEVBUFF_GEN_MASTER_SHARE_INPUT_SIZE,
            self.get_devbuff_base_address(1),
            self.DEVBUFF_SIZE,
            self.get_devbuff_base_address(2),
            self.DEVBUFF_SIZE,
            self.get_devbuff_base_address(3),
            self.DEVBUFF_SIZE,
        )

        if not oem_gen_master_share_res:
            raise SPSDKError(
                f"OEM generate master share command failed. Error: {self.mboot.status_string}\n"
                "Device probably needs reset due to doubled call of this command."
            )

        if (
            oem_gen_master_share_res[0] != self.DEVBUFF_GEN_MASTER_ENC_SHARE_OUTPUT_SIZE
            and oem_gen_master_share_res[1] != self.DEVBUFF_GEN_MASTER_ENC_MASTER_SHARE_OUTPUT_SIZE
            and oem_gen_master_share_res[2] != self.DEVBUFF_GEN_MASTER_CUST_CERT_PUK_OUTPUT_SIZE
        ):
            raise SPSDKError("OEM generate master share command has invalid results.")

        oem_enc_share = self.mboot.read_memory(
            self.get_devbuff_base_address(1),
            self.DEVBUFF_GEN_MASTER_ENC_SHARE_OUTPUT_SIZE,
        )
        if not oem_enc_share:
            raise SPSDKError(
                f"Cannot read OEM ENCRYPTED SHARE OUTPUT from device. Error: {self.mboot.status_string}"
            )
        self.store_temp_res("ENC_OEM_SHARE.bin", oem_enc_share)

        oem_enc_master_share = self.mboot.read_memory(
            self.get_devbuff_base_address(2),
            self.DEVBUFF_GEN_MASTER_ENC_MASTER_SHARE_OUTPUT_SIZE,
        )
        if not oem_enc_master_share:
            raise SPSDKError(
                f"Cannot read OEM ENCRYPTED MASTER SHARE OUTPUT from device. Error: {self.mboot.status_string}"
            )
        self.store_temp_res("ENC_OEM_MASTER_SHARE.bin", oem_enc_master_share)

        oem_cert = self.mboot.read_memory(
            self.get_devbuff_base_address(3),
            self.DEVBUFF_GEN_MASTER_CUST_CERT_PUK_OUTPUT_SIZE,
        )
        if not oem_cert:
            raise SPSDKError(
                f"Cannot read OEM CERTIFICATE from device. Error: {self.mboot.status_string}"
            )
        self.store_temp_res("OEM_CERT.bin", oem_cert)

        return oem_enc_share, oem_enc_master_share, oem_cert

    def generate_key(
        self, key_type: TrustProvOemKeyType, key_name: Optional[str] = None
    ) -> Tuple[bytes, bytes]:
        """Generate on device key pairs of provided type.

        :param key_type: Type of generated key pairs.
        :param key_name: optional name for storing temporary files.
        :raises SPSDKError: In case of any vulnerability.
        :return: Tuple with Private and Public key.
        """
        hsm_gen_key_res = self.mboot.tp_hsm_gen_key(
            key_type,
            0,
            self.devbuff_base,
            self.DEVBUFF_SIZE,
            self.get_devbuff_base_address(1),
            self.DEVBUFF_SIZE,
        )

        if not hsm_gen_key_res:
            raise SPSDKError(f"HSM generate key command failed. Error: {self.mboot.status_string}")

        if (
            hsm_gen_key_res[0] != self.DEVBUFF_HSM_GENKEY_KEYBLOB_SIZE
            and hsm_gen_key_res[1] != self.DEVBUFF_HSM_GENKEY_KEYBLOB_PUK_SIZE
        ):
            raise SPSDKError("OEM generate master share command has invalid results.")

        prk = self.mboot.read_memory(
            self.devbuff_base,
            self.DEVBUFF_HSM_GENKEY_KEYBLOB_SIZE,
        )
        if not prk:
            raise SPSDKError(
                f"Cannot read generated private key from device. Error: {self.mboot.status_string}"
            )

        puk = self.mboot.read_memory(
            self.get_devbuff_base_address(1),
            self.DEVBUFF_HSM_GENKEY_KEYBLOB_PUK_SIZE,
        )
        if not puk:
            raise SPSDKError(
                f"Cannot read generated public key from device. Error: {self.mboot.status_string}"
            )

        self.store_temp_res((key_name or str(TrustProvOemKeyType[str(key_type)])) + "_PRK.bin", prk)
        self.store_temp_res((key_name or str(TrustProvOemKeyType[str(key_type)])) + "_PUK.bin", puk)

        return prk, puk

    def wrap_key(self, cust_mk_sk: bytes) -> bytes:
        """Wrap the CUST_MK_SK key.

        :param cust_mk_sk : Customer Master Key Symmetric Key
        :raises SPSDKError: In case of any vulnerability.
        :return: Wrapped CUST_MK_SK by RFC3396.
        """
        if not self.mboot.write_memory(self.devbuff_base, cust_mk_sk):
            raise SPSDKError(
                f"Cannot write CUST_MK_SK into device. Error: {self.mboot.status_string}"
            )

        hsm_store_key_res = self.mboot.tp_hsm_store_key(
            TrustProvKeyType.CKDFK,
            0x01,
            self.devbuff_base,
            self.DEVBUFF_CUST_MK_SK_KEY_SIZE,
            self.get_devbuff_base_address(1),
            self.DEVBUFF_SIZE,
        )

        if not hsm_store_key_res:
            raise SPSDKError(f"HSM Store Key command failed. Error: {self.mboot.status_string}")

        if hsm_store_key_res[1] != self.DEVBUFF_WRAPPED_CUST_MK_SK_KEY_SIZE:
            raise SPSDKError("HSM Store Key command has invalid results.")

        wrapped_cust_mk_sk = self.mboot.read_memory(
            self.get_devbuff_base_address(1),
            self.DEVBUFF_WRAPPED_CUST_MK_SK_KEY_SIZE,
        )

        if not wrapped_cust_mk_sk:
            raise SPSDKError(
                f"Cannot read WRAPPED CUST_MK_SK from device. Error: {self.mboot.status_string}"
            )

        self.store_temp_res("CUST_MK_SK.bin", wrapped_cust_mk_sk)

        return wrapped_cust_mk_sk

    def sign_data_blob(self, data_to_sign: bytes, key: bytes) -> bytes:
        """Get HSM encryption sign for data blob.

        :param data_to_sign: Input data to sign.
        :param key: FW signing key (MFWISK).
        :raises SPSDKError: In case of any vulnerability.
        :return: Data blob signature (64 bytes).
        """
        if not self.mboot.write_memory(self.devbuff_base, key):
            raise SPSDKError(
                f"Cannot write signing key into device. Error: {self.mboot.status_string}"
            )
        if not self.mboot.write_memory(self.get_devbuff_base_address(1), data_to_sign):
            raise SPSDKError(
                f"Cannot write Data to sign into device. Error: {self.mboot.status_string}"
            )
        hsm_gen_key_res = self.mboot.tp_hsm_enc_sign(
            self.devbuff_base,
            len(key),
            self.get_devbuff_base_address(1),
            len(data_to_sign),
            self.get_devbuff_base_address(2),
            self.DEVBUFF_SB3_SIGNATURE_SIZE,
        )

        if hsm_gen_key_res != self.DEVBUFF_SB3_SIGNATURE_SIZE:
            raise SPSDKError(
                f"HSM signing command failed. Invalid signature size: {hsm_gen_key_res} "
                f"MBoot Status: {self.mboot.status_string}"
            )

        signature = self.mboot.read_memory(
            self.get_devbuff_base_address(2),
            self.DEVBUFF_SB3_SIGNATURE_SIZE,
        )
        if not signature:
            raise SPSDKError(
                f"Cannot read generated signature from device. Error: {self.mboot.status_string}"
            )

        self.store_temp_res("SB3_sign.bin", signature, "to_merge")

        return signature

    def store_temp_res(self, file_name: str, data: bytes, group: Optional[str] = None) -> None:
        """Storing temporary files into workspace.

        :param file_name: Name of file to store the data.
        :param data: Data to store.
        :param group: Subfolder name, defaults to None
        """
        if not self.workspace:
            return
        group_dir = os.path.join(self.workspace, group or "")
        if not os.path.isdir(group_dir):
            os.mkdir(group_dir)

        filename = os.path.join(self.workspace, group or "", file_name)
        write_file(data, filename, mode="wb")

    def get_cmd_from_config(self) -> List[Dict[str, Any]]:
        """Process command description into a command object.

        :return: Modified list of commands
        :raises SPSDKError: Unknown command
        """
        cfg_commands: List[Dict[str, Any]] = []
        if self.config_data and self.config_data.get("commands"):
            cfg_commands = self.config_data["commands"]
            for cmd in cfg_commands:
                cmd_cpy: Dict = cmd.copy()
                name, args = cmd_cpy.popitem()
                if name == "loadKeyBlob" and value_to_int(str(args["offset"])) == 0x04:
                    raise SPSDKError(
                        f"""The duplicated 'loadKeyBlob' on offset 0x04 from
                    configuration file:\n {args}."""
                    )

        return cfg_commands

    def encrypt_data_blocks(
        self, cust_fw_enc_key: bytes, sb3_header: bytes, data_cmd_blocks: List[bytes]
    ) -> List[bytes]:
        """Encrypt all data blocks on device.

        :param cust_fw_enc_key: Firmware encryption key.
        :param sb3_header: Un Encrypted SB3 file header.
        :param data_cmd_blocks: List of un-encrypted SB3 file command blocks.
        :raises SPSDKError: In case of any vulnerability.
        :return: List of encrypted command blocks on device.
        """
        if not self.mboot.write_memory(self.devbuff_base, cust_fw_enc_key):
            raise SPSDKError(
                f"Cannot write customer fw encryption key into device. Error: {self.mboot.status_string}"
            )
        self.store_temp_res("SB3_header.bin", sb3_header, "to_encrypt")
        if not self.mboot.write_memory(self.get_devbuff_base_address(1), sb3_header):
            raise SPSDKError(
                f"Cannot write SB3 header into device. Error: {self.mboot.status_string}"
            )

        encrypted_blocks = []
        for data_cmd_block_ix, data_cmd_block in enumerate(data_cmd_blocks, start=1):
            self.store_temp_res(f"SB3_block_{data_cmd_block_ix}.bin", data_cmd_block, "to_encrypt")
            if not self.mboot.write_memory(self.get_devbuff_base_address(2), data_cmd_block):
                raise SPSDKError(
                    f"Cannot write SB3 data block{data_cmd_block_ix} into device. "
                    f"Error: {self.mboot.status_string}"
                )
            key_id = CmdLoadKeyBlob.get_key_id(
                self.family, CmdLoadKeyBlob.KeyTypes.NXP_CUST_KEK_INT_SK
            )
            if not self.mboot.tp_hsm_enc_blk(
                self.devbuff_base,
                len(cust_fw_enc_key),
                key_id,
                self.get_devbuff_base_address(1),
                len(sb3_header),
                data_cmd_block_ix,
                self.get_devbuff_base_address(2),
                self.DEVBUFF_DATA_BLOCK_SIZE,
            ):
                raise SPSDKError(
                    f"Cannot run SB3 data block_{data_cmd_block_ix} HSM Encryption in device. "
                    f"Error: {self.mboot.status_string}"
                )

            encrypted_block = self.mboot.read_memory(
                self.get_devbuff_base_address(2),
                self.DEVBUFF_DATA_BLOCK_SIZE,
            )
            if not encrypted_block:
                raise SPSDKError(
                    f"Cannot read SB3 data block_{data_cmd_block_ix} from device. "
                    f"Error: {self.mboot.status_string}"
                )

            self.store_temp_res(f"SB3_block_{data_cmd_block_ix}.bin", encrypted_block, "encrypted")

            encrypted_blocks.append(encrypted_block)

        return encrypted_blocks


def get_cust_mk_sk(key: str) -> bytes:
    """Get binary from text or binary file.

    :param key: Binary customer master key symmetric key file.
    :return: Binary array loaded from file.
    :raises SPSDKValueError: When invalid input value is recognized.
    """
    cust_mk_sk = load_binary(key)

    if len(cust_mk_sk) != 32:
        raise SPSDKValueError(
            f"Invalid length of CUST_MK_SK INPUT ({len(cust_mk_sk )} not equal to 32)."
        )

    return cust_mk_sk


def get_oem_share_input(binary: str) -> bytes:
    """Get binary from text or binary file.

    :param binary: Path to binary file.
    :return: Binary array loaded from file.
    :raises SPSDKValueError: When invalid input value is recognized.
    """
    if binary:
        oem_share_input = load_binary(binary)
    else:
        oem_share_input = crypto_backend().random_bytes(16)

    if len(oem_share_input) != 16:
        raise SPSDKValueError(
            f"Invalid length of OEM SHARE INPUT ({len(oem_share_input)} not equal to 16)."
        )

    return oem_share_input


@click.group(name="nxpdevhsm", no_args_is_help=True, cls=CommandsTreeGroup)
@spsdk_apps_common_options
def main(log_level: int) -> int:
    """Nxpdevhsm application is designed to create SB3 provisioning file for initial provisioning of device by OEM."""
    spsdk_logger.install(level=log_level)
    return 0


@main.command(no_args_is_help=True)
@isp_interfaces(uart=True, usb=True, sdio=True, lpcusbsio=True, buspal=True, json_option=False)
@click.option(
    "-k",
    "--key",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    required=True,
    help="""Customer Master Key Symmetric Key secret file (32-bytes long binary file).
    CUST_MK_SK (provisioned by OEM, known by OEM).
    This is a 256-bit pre-shared AES key provisioned by OEM. CUST_MK_SK is used to derive FW image encryption keys.""",
)
@click.option(
    "-o",
    "--oem-share-input",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    help="OEM share input file to use as a seed to randomize the provisioning process (16-bytes long binary file).",
)
@click.option(
    "-w",
    "--workspace",
    type=click.Path(file_okay=False),
    required=False,
    help="Workspace folder to store temporary files, that could be used for future review.",
)
@click.option(
    "-j",
    "--container-conf",
    type=click.Path(exists=True),
    required=False,
    help="""JSON/YAML container configuration file to produce secure binary v3.x.
    This configuration file must contain family, list of commands and description""",
)
@click.option(
    "-f",
    "--family",
    required=True,
    help="Select the chip family.",
    type=click.Choice(DeviceHsm.get_supported_families(), case_sensitive=False),
)
@click.option(
    "-ir/-IR",
    "--initial-reset/--no-init-reset",
    default=False,
    help=(
        "Reset device BEFORE DevHSM operation. The DevHSM operation can run only once between resets. "
        "Do not enable this option on Linux/Mac when using USB. By default this reset is DISABLED."
    ),
)
@click.option(
    "-fr/-FR",
    "--final-reset/--no-final-reset",
    default=True,
    help=(
        "Reset device AFTER DevHSM operation. This reset is required if you need to use the device "
        "after DevHSM operation for other security related operations (e.g. receive-sb-file). "
        "By default this reset is ENABLED."
    ),
)
@click.argument("output-path", type=click.Path())
def generate(
    port: str,
    usb: str,
    sdio: str,
    buspal: str,
    lpcusbsio: str,
    oem_share_input: str,
    key: str,
    output_path: str,
    workspace: str,
    container_conf: str,
    timeout: int,
    family: str,
    initial_reset: bool,
    final_reset: bool,
) -> None:
    """Generate provisioning SB3.1 file.

    \b
    PATH    - output file path, where the final provisioned SB file will be stored.
    """
    interface = get_interface(
        module="mboot", port=port, usb=usb, lpcusbsio=lpcusbsio, timeout=timeout, buspal=buspal
    )
    assert isinstance(interface, MBootInterface)

    oem_share_in = get_oem_share_input(oem_share_input)
    cust_mk_sk = get_cust_mk_sk(key)
    if container_conf:
        family_from_cfg = load_configuration(container_conf).get("family")
        if not isinstance(family_from_cfg, str):
            raise SPSDKError("Family parameter is not provided in the container configuration")
    if (family and container_conf) and (family != family_from_cfg):
        raise SPSDKError(
            f"Family from json configuration file: {family_from_cfg} differs from the family parameter {family}"
        )

    with McuBoot(interface) as mboot:
        devhsm = DeviceHsm(
            mboot=mboot,
            cust_mk_sk=cust_mk_sk,
            oem_share_input=oem_share_in,
            info_print=click.echo,
            container_conf=container_conf,
            workspace=workspace,
            family=family,
            initial_reset=initial_reset,
            final_reset=final_reset,
        )
        devhsm.create_sb3()
        write_file(devhsm.export(), output_path, "wb")

    click.echo(f"Final SB3 file has been written: {os.path.abspath(output_path)}")


@main.command(name="get-template", no_args_is_help=True)
@click.option(
    "-f",
    "--family",
    required=True,
    help="Select the chip family.",
    type=click.Choice(DeviceHsm.get_supported_families(), case_sensitive=False),
)
@click.option(
    "-o",
    "--overwrite",
    default=False,
    type=bool,
    is_flag=True,
    help="Allow overwriting existing template file.",
)
@click.argument("output", type=click.Path(resolve_path=True))
def get_template_command(family: str, overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format.

    The template file name is specified as argument of this command.
    """
    get_template(family, overwrite, output)


def get_template(family: str, overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format."""
    if not os.path.isfile(output) or overwrite:
        click.echo(f"Creating {output} template file.")
        write_file(DeviceHsm.generate_config_template(family)[f"sb31_{family}_devhsm"], output)
    else:
        click.echo(f"Skip creating {output}, this file already exists.")


@catch_spsdk_error
def safe_main() -> None:
    """Call the main function."""
    sys.exit(main())  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    safe_main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Trust Provisioning HOST application support."""
import base64
import concurrent.futures
import logging
import math
import multiprocessing
import os
import secrets
import struct
import time
from functools import partial
from typing import Callable, Optional, Sequence

from spsdk import SPSDKError
from spsdk.crypto import EllipticCurvePublicKey, serialization
from spsdk.crypto.loaders import Encoding, extract_public_key, load_pem_public_key
from spsdk.tp.data_container import AuditLog, DataEntry, PayloadType
from spsdk.utils.database import Database
from spsdk.utils.misc import Timeout, value_to_int, write_file

from . import TP_DATA_FOLDER, SPSDKTpError, TpDevInterface, TpTargetInterface
from .adapters.tptarget_blhost import TpTargetBlHost
from .adapters.utils import detect_new_usb_path, get_current_usb_paths, update_usb_path
from .data_container import AuditLogCounter, AuditLogRecord, Container

logger = logging.getLogger(__name__)

REOPEN_WAIT_TIME = 0.3
ALLOW_ARBITRARY_START = True


class TrustProvisioningHost:
    """Trust provisioning support in none trusted environment."""

    def __init__(
        self,
        tpdev: TpDevInterface,
        tptarget: TpTargetInterface,
        info_print: Callable[[str], None],
    ) -> None:
        """Trust Provisioning Host support class.

        :param tpdev: TP device instance
        :param tptarget: TP target instance
        :param info_print: Method for printing messages
        """
        self.tpdev = tpdev
        self.tptarget = tptarget
        self.info_print = info_print

    def load_provisioning_fw(
        self,
        prov_fw: bytes,
        family: str,
        timeout: int = 60,
        database: Optional[Database] = None,
        skip_test: bool = True,
        keep_target_open: bool = True,
        skip_usb_enumeration: bool = False,
    ) -> None:
        """Method loads the provisioning firmware into device.

        :param prov_fw: Provisioning Firmware data
        :param family: Chip family
        :param timeout: Timeout for loading provisioning firmware operation in seconds.
        :param database: Database of all supported devices (automatic lookup if not specified)
        :param skip_test: Skip test for checking that OEM Provisioning Firmware booted-up
        :param keep_target_open: Keep target device open
        :param skip_usb_enumeration: Skip USB enumeration after loading the Provisioning firmware
        :raises SPSDKTpError: The Provisioning firmware doesn't boot
        """
        if database is None:
            logger.debug("Looking up device in database")
            database = Database(os.path.join(TP_DATA_FOLDER, "database.yaml"))
        if family not in database.devices.device_names:
            raise SPSDKTpError(f"Database info missing for '{family}'")
        try:
            if not self.tptarget.is_open:
                self.tptarget.open()
            self.info_print("1.1.Step - Updating CFPA page")
            self.update_cfpa_page(family=family, database=database)
            self.info_print("1.2.Step - Erase memory for provisioning firmware")
            self.erase_memory(family=family, database=database)
            self.info_print("1.3.Step - Loading OEM provisioning firmware")
        except SPSDKError as e:
            self.tptarget.close()
            raise SPSDKTpError(f"Unable to prepare the MCU: {e}") from e

        try:
            if self.tptarget.uses_usb:
                initial_usb_set = get_current_usb_paths()

            self.tptarget.load_sb_file(prov_fw, timeout)
            # Need to reset the connection due to re-init on the MCU side
            self.tptarget.close()

            if self.tptarget.uses_usb and not skip_usb_enumeration:
                assert isinstance(self.tptarget, TpTargetBlHost)
                new_usb_path = detect_new_usb_path(initial_set=initial_usb_set)
                update_usb_path(self.tptarget, new_usb_path=new_usb_path)

            logger.info(f"Waiting for {REOPEN_WAIT_TIME} seconds for the ProvFW to boot up.")
            time.sleep(REOPEN_WAIT_TIME)

            if not skip_test:
                self.info_print("1.4.Step - Checking whether provisioning firmware booted.")
                self.tptarget.open()
                if not self.tptarget.check_provisioning_firmware():
                    raise SPSDKError("Provisioning firmware did not boot properly")

            if keep_target_open and not self.tptarget.is_open:
                self.tptarget.open()

        except SPSDKError as e:
            self.tptarget.close()
            raise SPSDKTpError(
                f"Can't load/connect to the TrustProvisioning Firmware. Error: {e}\n"
                "Please make sure your device supports TrustProvisioning."
            ) from e

    def update_cfpa_page(self, family: str, database: Database) -> None:
        """Update CFPA page according to chip family."""
        if not database.get_device_value("need_cfpa_update", family):
            logger.info("CFPA update not required")
            return

        cfpa_address = value_to_int(database.get_device_value("cfpa_address", family))
        cfpa_size = value_to_int(database.get_device_value("cfpa_size", family))
        cfpa_version_offset = value_to_int(database.get_device_value("version_offset", family))
        cfpa_revoke_offset = value_to_int(database.get_device_value("revoke_offset", family))

        # change bytes to bytearray to make it writeable
        cfpa_data = bytearray(self.tptarget.read_memory(cfpa_address, cfpa_size))

        # CFPA REVOKE field update
        if database.get_device_value("need_revoke_update", family):
            cfpa_revoke: int = struct.unpack_from("<L", cfpa_data, offset=cfpa_revoke_offset)[0]
            if cfpa_revoke & 0x55 == 0x55:
                logger.info("RKTH_REVOKE is already set, no need to update CFPA")
                return
            # just set required bits (in case user already set other bits)
            cfpa_revoke |= 0x55
            struct.pack_into("<L", cfpa_data, cfpa_revoke_offset, cfpa_revoke)
        else:
            raise SPSDKTpError(f"Don't know how to update CFPA for {family}")

        # cfpa update is a generic thing for all families
        cfpa_version: int = struct.unpack_from("<L", cfpa_data, offset=cfpa_version_offset)[0]
        cfpa_version += 1
        struct.pack_into("<L", cfpa_data, cfpa_version_offset, cfpa_version)

        self.tptarget.write_memory(cfpa_address, data=bytes(cfpa_data))
        logger.info("CFPA update completed")

    def erase_memory(self, family: str, database: Database) -> None:
        """Erase part(s) of flash memory if needed."""
        if not database.get_device_value("erase_memory", family):
            logger.info("Erasing memory is not needed")
            return

        start = value_to_int(database.get_device_value("erase_memory_start", family))
        length = value_to_int(database.get_device_value("erase_memory_length", family))

        self.tptarget.erase_memory(address=start, length=length)
        logger.info("Erasing memory completed")

    def do_provisioning(
        self,
        family: str,
        audit_log: str,
        prov_fw: Optional[bytes] = None,
        product_fw: Optional[bytes] = None,
        timeout: int = 60,
        save_debug_data: bool = False,
    ) -> None:
        """Do provisioning process.

        :param family: Chip family
        :param audit_log: Path to audit log
        :param prov_fw: Use own provisioning firmware, defaults to None
        :param product_fw: Load also the final product application, defaults to None
        :param timeout: The timeout of operation is seconds.
        :param save_debug_data: Save transmitted data in CWD for debugging purposes
        :raises SPSDKTpError: Device family is not supported
        :raises SPSDKTpError: Error during trust-provisioning operation
        """
        try:
            loc_timeout = Timeout(timeout, "s")

            logger.debug("Looking up device in database")
            database = Database(os.path.join(TP_DATA_FOLDER, "database.yaml"))
            if family not in database.devices.device_names:
                raise SPSDKTpError(f"Database info missing for '{family}'")

            logger.debug("Opening TP DEVICE interface")
            self.tpdev.open()
            audit_log_dirname = os.path.dirname(os.path.abspath(audit_log))

            if os.path.isfile(audit_log):
                self.info_print("0.Step - Check Audit Log ownership")
                self.tpdev.check_log_owner(audit_log)

            elif not os.path.exists(audit_log_dirname):
                self.info_print(" - Creating directory for the audit log")
                os.makedirs(audit_log_dirname)

            logger.debug("Opening TP TARGET interface")
            self.tptarget.open()

            self.info_print("1.Step - Provide to target provisioning firmware if needed.")
            if prov_fw:
                self.load_provisioning_fw(
                    prov_fw=prov_fw,
                    family=family,
                    timeout=loc_timeout.get_rest_time_ms(True),
                    database=database,
                    skip_test=True,
                    keep_target_open=True,
                    skip_usb_enumeration=False,
                )

            self.info_print("2.Step - Get the initial challenge from TP device.")
            challenge = self.tpdev.get_challenge(timeout=loc_timeout.get_rest_time_ms(True))

            logger.info(f"TP Challenge:\n{Container.parse(challenge)}")
            if save_debug_data:
                write_file(challenge, "x_challenge.bin", "wb")

            self.info_print("3.Step - Prove a genuinity in TP target.")
            tp_data = self.tptarget.prove_genuinity_challenge(
                challenge, timeout=loc_timeout.get_rest_time_ms(True)
            )

            logger.info(f"TP Response:\n{Container.parse(tp_data)}")
            if save_debug_data:
                write_file(tp_data, "x_tp_response.bin", "wb")

            self.info_print("4.Step - Authenticate TP response from TP target.")
            wrapped_data = self.tpdev.authenticate_response(
                tp_data, timeout=loc_timeout.get_rest_time_ms(True)
            )

            logger.info(f"TP ISP WRAPPED DATA:\n{Container.parse(wrapped_data)}")
            if save_debug_data:
                write_file(wrapped_data, "x_wrapped_data.bin", "wb")

            self.info_print("5.Step - Create Audit Log record.")
            self.create_audit_log_record(wrapped_data, audit_log)

            self.info_print("6.Step - Set the wrapped data from the TP device to target.")
            self.tptarget.set_wrapped_data(wrapped_data, timeout=loc_timeout.get_rest_time_ms(True))

            self.info_print("7.Step - The target is provisioned, commencing Reset.")
            self.tptarget.reset_device()

            if product_fw:
                self.info_print("8.Step - Loading customer application.")
                logger.info(f"Waiting for {REOPEN_WAIT_TIME} seconds for the ROM to boot up.")
                time.sleep(REOPEN_WAIT_TIME)
                self.tptarget.open()
                self.tptarget.load_sb_file(product_fw, timeout=loc_timeout.get_rest_time_ms(True))

            self.info_print(
                f"Trust provisioning process ends correctly in {loc_timeout.get_consumed_time_ms()} ms."
            )
        except:
            self.info_print(
                f"Trust provisioning process FAILED in {loc_timeout.get_consumed_time_ms()} ms."
            )
            raise
        finally:
            self.tpdev.close()
            self.tptarget.close()

    def create_audit_log_record(self, data: bytes, audit_log: str) -> None:
        """Create an audit log record out of data representing ISP_WRAP_DATA container."""
        logger.info(f"Using log file {audit_log}")
        record = AuditLogRecord.from_data(container_data=data)
        record.save(audit_log, str(self.tpdev.descriptor.get_id()))

    @staticmethod
    def verify_extract_log(
        audit_log: str,
        audit_log_key: str,
        destination: Optional[str] = None,
        skip_nxp: bool = False,
        skip_oem: bool = False,
        cert_index: Optional[int] = None,
        encoding: Encoding = Encoding.PEM,
        max_processes: Optional[int] = None,
        info_print: Callable[[str], None] = lambda x: None,
        force_rewrite: bool = False,
    ) -> AuditLogCounter:
        """Verifying audit log with given key (public/private).

        :param audit_log: Path to audit log
        :param audit_log_key: Path to public/private key for verification
        :param destination: Path to destination directory for extracted certificates
        :param skip_nxp: Skip extracting the NXP Devattest certificates
        :param skip_oem: Skip extracting the OEM x509 Devattest certificates
        :param cert_index: Select single OEM certificate to extract, default None = all certificates
        :param encoding: Certificate encoding, defaults to Encoding.PEM
        :param max_processes: Maximum number od parallel process to use, defaults to CPU count
        :param info_print: Method for printing messages
        :param force_rewrite: Skip checking for empty destination directory and rewrite existing content
        :raises SPSDKTpError: Audit log record or chain is invalid
        """
        try:
            loc_timeout = Timeout(timeout=0, units="s")
            info_print(f"Verifying the Audit log: {audit_log}")
            logger.info(f"Extracting public key from {audit_log_key}")

            log_key = extract_public_key(audit_log_key)
            # PublicKey can't be passed to other processes we have serialize it
            log_key_pem_data = log_key.public_bytes(
                encoding=Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo
            )
            assert isinstance(log_key, EllipticCurvePublicKey)

            if destination:
                os.makedirs(destination, exist_ok=True)

            # create partial method to save typing later, because readability counts
            store_certificate_method = partial(
                _extract_certificates,
                destination_dir=destination,
                skip_nxp=skip_nxp,
                skip_oem=skip_oem,
                cert_index=cert_index,
                encoding=encoding,
                force_rewrite=force_rewrite,
            )

            logger.info("Start loading audit log")
            log_record_count = AuditLog.record_count(audit_log)
            info_print(f"Found {log_record_count} record(s) in the audit log.")
            _, first_record = next(AuditLog.records(audit_log))
            if first_record.prod_counter_int != 1:
                logger.warning(
                    f"First record in audit log has PROV_COUNTER = {first_record.prod_counter_int} (Expecting 1)"
                )

            logger.info("Start verifying")
            verify_time = Timeout(timeout=0)
            summary_counter = AuditLogCounter()
            if log_record_count < 100 or max_processes == 1:
                logger.info("Using own process for verification")
                summary_counter = _verify_extract_chain(
                    audit_log=audit_log,
                    log_slice=slice(0, log_record_count),
                    public_key_pem_data=log_key_pem_data,
                    store_cert_method=store_certificate_method,
                )
            else:
                # using parallel execution
                process_count = max_processes or multiprocessing.cpu_count()
                logger.info(f"Using {process_count} processes for verification")
                log_slices = _get_log_slices(log_record_count, process_count=process_count)
                with concurrent.futures.ProcessPoolExecutor(max_workers=process_count) as executor:
                    future_to_slice = {
                        executor.submit(
                            _verify_extract_chain,
                            audit_log=audit_log,
                            log_slice=log_slice,
                            public_key_pem_data=log_key_pem_data,
                            store_cert_method=store_certificate_method,
                        ): log_slice
                        for log_slice in log_slices
                    }
                    for future in concurrent.futures.as_completed(future_to_slice):
                        summary_counter += future.result()
            logger.info(f"Verification completed in {verify_time.get_consumed_time_ms()} ms.")
            info_print(
                f"Audit log verification successfully finished in {loc_timeout.get_consumed_time_ms()} ms."
            )
            info_print(str(summary_counter))
            return summary_counter
        except:
            info_print(f"Audit log verification FAILED in {loc_timeout.get_consumed_time_ms()} ms.")
            raise

    def get_counters(self, timeout: int = 60) -> None:
        """Seal the provisioning device."""
        try:
            loc_timeout = Timeout(timeout=timeout, units="s")
            logger.debug("Opening TP device")
            self.tpdev.open()

            self.info_print("Retrieving counters from TP Device.")

            self.info_print(f"Current provisioning counter: {self.tpdev.get_prov_counter():,}")
            self.info_print(f"Provisioning attempts left  : {self.tpdev.get_prov_remainder():,}")
            self.info_print(
                f"TP device counters retrieval ended correctly in {loc_timeout.get_consumed_time_ms()} ms."
            )
        except:
            self.info_print(
                f"TP device counters retrieval FAILED in {loc_timeout.get_consumed_time_ms()} ms."
            )
            raise
        finally:
            logger.debug("Closing TP device")
            self.tpdev.close()

    def check_audit_log_owner(self, audit_log: str, timeout: int = 60) -> None:
        """Check if this TP Device's ID is present in the audit log."""
        try:
            loc_timeout = Timeout(timeout=timeout, units="s")
            logger.debug("Opening TP device")
            self.tpdev.open()

            self.info_print(f"Checking ownership of audit log: {audit_log}")
            tp_device_id = self.tpdev.descriptor.get_id()
            logger.debug(f"TP Device ID: {tp_device_id}")
            audit_log_id = AuditLog.properties(audit_log).tp_device_id
            logger.debug(f"Audit log ID: {audit_log_id}")
            if tp_device_id != audit_log_id:
                raise SPSDKTpError(
                    f"TP Device ID and Audit Log ID differ! "
                    f"TP Device ID: {tp_device_id}, Audit Log ID: {audit_log_id}"
                )
            self.info_print(f"TP Device ID and Audit Log ID are the same: {tp_device_id}")
            self.info_print(
                f"Checking ownership of audit log ended correctly in {loc_timeout.get_consumed_time_ms()} ms."
            )
        except:
            self.info_print(
                f"Checking ownership of audit log FAILED in {loc_timeout.get_consumed_time_ms()} ms."
            )
            raise
        finally:
            logger.debug("Closing TP device")
            self.tpdev.close()

    def get_tp_response(
        self,
        response_file: str,
        timeout: int = 60,
        challenge: Optional[bytes] = None,
        oem_id_key_count: int = 0,
    ) -> None:
        """Retrieve TP_RESPONSE from the target.

        :param response_file: Path where to store TP_RESPONSE
        :param timeout: The timeout of operation is seconds, defaults to 60
        :param challenge: Challenge for the response, defaults to None
        :param oem_id_key_count: Number of OEM keys to include in the response, defaults to 0
        :raises SPSDKTpError: Failure to retrieve the response
        """
        try:
            loc_timeout = Timeout(timeout=timeout, units="s")
            logger.debug("Opening TP target")
            self.tptarget.open()

            challenge = challenge or secrets.token_bytes(16)
            if len(challenge) != 16:
                raise SPSDKTpError("Challenge has to be 16B long")

            challenge_container = Container()
            challenge_container.add_entry(
                DataEntry(
                    payload=challenge,
                    payload_type=PayloadType.NXP_EPH_CHALLENGE_DATA_RND,
                    extra=oem_id_key_count << 2,
                )
            )
            logger.info(f"TP Challenge:\n{challenge_container}")

            tp_response = self.tptarget.prove_genuinity_challenge(
                challenge=challenge_container.export(),
                timeout=loc_timeout.get_rest_time_ms(raise_exc=True),
            )
            logger.info(f"TP Response:\n{Container.parse(tp_response)}")

            write_file(data=tp_response, path=response_file, mode="wb")
            self.info_print(f"TP_RESPONSE stored to {response_file}.")
            self.info_print(
                f"Retrieving TP_RESPONSE ends correctly in {loc_timeout.get_consumed_time_ms()} ms."
            )
        except:
            self.info_print(
                f"Retrieving TP_RESPONSE failed in {loc_timeout.get_consumed_time_ms()} ms."
            )
            raise
        finally:
            logger.debug("Closing TP target")
            self.tptarget.close()


def _get_log_slices(log_length: int, process_count: int) -> Sequence[slice]:
    """Create list of slices for slicing the audit log depending on process count."""
    chunk_size = math.ceil(log_length / process_count)
    slices = [slice(i, i + chunk_size) for i in range(0, log_length, chunk_size)]
    logger.debug(f"Audit log split into slices: {slices}")
    return slices


def _verify_extract_chain(
    audit_log: str,
    public_key_pem_data: bytes,
    store_cert_method: Callable[[AuditLogRecord], AuditLogCounter],
    log_slice: slice,
) -> AuditLogCounter:
    """Verify content of AuditLog and optionally store certificates."""
    public_key = load_pem_public_key(public_key_pem_data)
    assert isinstance(public_key, EllipticCurvePublicKey)
    counter = AuditLogCounter()

    # if the current slice is not at the begging of file
    # we have to fetch previous record for chain verification
    # if the current slice is the first one in log, the starting hash should be 0
    start = log_slice.start
    fetch_previous = False
    if start != 0:
        start -= 1
        fetch_previous = True

    records = AuditLog.records(audit_log, id_slice=(start, log_slice.stop))
    if fetch_previous:
        previous_hash = next(records)[1].new_hash()
    else:
        previous_hash = bytes(32)

    for i, record in records:
        if not record.is_valid(public_key):
            raise SPSDKTpError(f"Log entry #{i} has an invalid signature!")
        if record.start_hash != previous_hash:
            if ALLOW_ARBITRARY_START and (i == 1):
                # TODO: need to get confirmation on this one
                pass
            else:
                raise SPSDKTpError(f"Audit log chain is broken between records #{i - 1} - #{i}")
        previous_hash = record.new_hash()
        counter.check_count += 1
        counter += store_cert_method(record)
    return counter


def _extract_certificates(
    log_record: AuditLogRecord,
    destination_dir: Optional[str] = None,
    skip_nxp: bool = False,
    skip_oem: bool = False,
    cert_index: Optional[int] = None,
    encoding: Encoding = Encoding.PEM,
    force_rewrite: bool = False,
) -> AuditLogCounter:
    """Extract certificates from the audit log into destination_dir.

    :param audit_log: Path to audit log
    :param destination_dir: Path to destination directory
    :param skip_nxp: Skip extracting the NXP Devattest certificates
    :param skip_oem: Skip extracting the OEM x509 Devattest certificates
    :param info_print: Method for printing messages
    :param encoding: Certificate encoding, defaults to Encoding.PEM
    :raises SPSDKTpError: Destination directory doesn't exist
    :return: Number of generated NXP and OEM certificates
    """
    counter = AuditLogCounter()
    if not destination_dir:
        return counter
    if not os.path.isdir(destination_dir):
        raise SPSDKTpError(f"Directory {destination_dir} doesn't exist.")
    if not skip_nxp:
        nxp_id_name = f"device_{log_record.prod_counter_int:06}_identity_cert.bin"
        _write_cert_data(log_record.nxp_id_cert, nxp_id_name, destination_dir, force_rewrite)
        counter.nxp_count += 1
    if not skip_oem:
        for idx, oem_cert in enumerate(log_record.oem_id_certs):
            if cert_index and cert_index != idx:
                continue
            if not oem_cert:
                continue
            oem_id_name = f"device_{log_record.prod_counter_int:06}_device_cert_{idx}.cer"
            cert_data = _encode_cert_data(cert_data=oem_cert, encoding=encoding)
            _write_cert_data(cert_data, oem_id_name, destination_dir, force_rewrite)
            counter.oem_count += 1
    return counter


def _write_cert_data(cert_data: bytes, name: str, path: str, force_rewrite: bool) -> None:
    """Write certificate data into a file.

    :param cert_data: Certificate data
    :param name: Name of the certificate file
    :param path: Destination directory
    :param force_rewrite: Rewrite existing certificates
    :raises SPSDKTpError: Certificate already exists and force_rewrite is disabled
    """
    logger.debug(f"Writing {name}")
    destination = os.path.join(path, name)
    if os.path.exists(destination) and not force_rewrite:
        raise SPSDKTpError(
            f"{destination} already exists! To rewrite existing files use --force-rewrite flag."
        )
    write_file(cert_data, destination, mode="wb")


def _encode_cert_data(cert_data: bytes, encoding: Encoding = Encoding.PEM) -> bytes:
    """Encode certificate data with given encoding.

    :param cert_data: Certificate binary data (DER)
    :param encoding: Output encoding, defaults to Encoding.PEM
    :raises SPSDKTpError: Unsupported encoding
    :return: Encoded certificate data
    """
    if encoding == Encoding.DER:
        return cert_data
    if encoding == Encoding.PEM:
        data = base64.b64encode(cert_data)
        lines = [data[i : i + 64] for i in range(0, len(data), 64)]
        lines.insert(0, b"-----BEGIN CERTIFICATE-----")
        lines.append(b"-----END CERTIFICATE-----")
        return b"\n".join(lines)
    raise SPSDKTpError(f"Unsupported encoding: {encoding}")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""CLI application for various cryptographic operations."""

import copy
import hashlib
import os
import sys
from typing import Dict, Union

import click
from Crypto.PublicKey import ECC, RSA

from spsdk import SPSDKError
from spsdk.apps.nxpcertgen import main as cert_gen_main
from spsdk.apps.nxpkeygen import main as key_gen_main
from spsdk.apps.utils import spsdk_logger
from spsdk.apps.utils.common_cli_options import (
    CommandsTreeGroup,
    GroupAliasedGetCfgTemplate,
    spsdk_apps_common_options,
)
from spsdk.apps.utils.utils import SPSDKAppError, catch_spsdk_error
from spsdk.utils.misc import load_binary, write_file


@click.group(name="nxpcrypto", no_args_is_help=True, cls=CommandsTreeGroup)
@spsdk_apps_common_options
def main(log_level: int) -> None:
    """Collection of utilities for cryptographic operations."""
    spsdk_logger.install(level=log_level)


@main.command(name="digest", no_args_is_help=True)
@click.option(
    "-h",
    "--hash",
    "hash_name",
    required=True,
    type=click.Choice(list(hashlib.algorithms_available), case_sensitive=False),
    help="Name of a hash to use.",
)
@click.option(
    "-i",
    "--infile",
    type=click.Path(exists=True, dir_okay=False),
    required=True,
    help="Path to a file to digest.",
)
@click.option(
    "-c",
    "--compare",
    metavar="PATH | DIGEST",
    help="Reference digest to compare. It may be directly on the command line or fetched from a file.",
)
def digest(hash_name: str, infile: str, compare: str) -> None:
    """Computes digest/hash of the given file."""
    data = load_binary(infile)
    hasher = hashlib.new(hash_name.lower())
    hasher.update(data)
    hexdigest = hasher.hexdigest()
    click.echo(f"{hash_name.upper()}({infile})= {hexdigest}")
    if compare:
        # assume comparing to a file
        if os.path.isfile(compare):
            with open(compare) as f:
                compare_data = f.readline().strip()
                # assume format generated by openssl
                if "=" in compare_data:
                    ref = compare_data.split("=")[-1].strip()
                # assume hash is on the fist line
                else:
                    ref = compare_data
        else:
            ref = compare
        if ref.lower() == hexdigest.lower():
            click.echo("Digests are the same.")
        else:
            raise SPSDKAppError("Digests differ!")


@main.group(name="cert", no_args_is_help=True, cls=GroupAliasedGetCfgTemplate)
def cert() -> None:
    """Group of command for working with x509 certificates."""


cert.add_command(cert_gen_main.commands["generate"], name="generate")
cert.add_command(cert_gen_main.commands["get-template"], name="get-template")
cert.add_command(cert_gen_main.commands["verify"], name="verify")


@main.group(name="key", no_args_is_help=True)
def key_group() -> None:
    """Group of commands for working with asymmetric keys."""


key_gen_copy = copy.deepcopy(key_gen_main)
key_gen_copy.name = "generate"
key_group.add_command(key_gen_copy)


@key_group.command(name="convert", no_args_is_help=True)
@click.option(
    "-f",
    "--output-format",
    type=click.Choice(["PEM", "DER", "RAW"], case_sensitive=False),
    help="Desired output format.",
)
@click.option(
    "-i",
    "--infile",
    type=click.Path(exists=True, dir_okay=False),
    required=True,
    help="Path to key file to convert.",
)
@click.option(
    "-o",
    "--outfile",
    type=click.Path(dir_okay=False),
    help="Path to output file.",
)
@click.option(
    "-p",
    "--puk",
    is_flag=True,
    default=False,
    help="Extract public key instead of converting private key.",
)
@click.option(
    "--use-pkcs8/--no-pkcs8",
    is_flag=True,
    default=True,
    help="Use/don't use PKCS8 encoding for private keys, default: --use-pkcs8",
)
def convert(output_format: str, infile: str, outfile: str, puk: bool, use_pkcs8: bool) -> None:
    """Convert Asymmetric key into various formats."""
    key_data = load_binary(infile)
    key = reconstruct_key(key_data=key_data)
    if puk:
        key = key.public_key()

    if output_format.upper() in ["PEM", "DER"]:
        params: Dict[str, Union[str, int, bool]] = {
            "format": output_format.upper(),
        }
        if key.has_private():
            if isinstance(key, ECC.EccKey):
                params["use_pkcs8"] = use_pkcs8
            if isinstance(key, RSA.RsaKey):
                params["pkcs"] = 8 if use_pkcs8 else 1
        tmp_data = key.export_key(**params)  # type: ignore  # yeah, MyPy doesn't like dict unpacking
        out_data = tmp_data.encode("utf-8") if isinstance(tmp_data, str) else tmp_data
    if output_format.upper() == "RAW":
        if not isinstance(key, ECC.EccKey):
            raise SPSDKError("Converting to RAW is supported only for ECC keys")
        key_size = key.pointQ.size_in_bytes()
        if key.has_private():
            out_data = key.d.to_bytes(key_size)  # type: ignore  # this `to_bytes` doesn't have byteorder
        else:
            x = key.pointQ.x.to_bytes(key_size)  # type: ignore  # this `to_bytes` doesn't have byteorder
            y = key.pointQ.y.to_bytes(key_size)  # type: ignore  # this `to_bytes` doesn't have byteorder
            out_data = x + y

    write_file(out_data, outfile, mode="wb")


@key_group.command(name="verify", no_args_is_help=True)
@click.argument("key1", type=click.Path(exists=True, dir_okay=False))
@click.argument("key2", type=click.Path(exists=True, dir_okay=False))
def check_keys(key1: str, key2: str) -> None:
    """Check whether provided keys form a key pair or represent the same key."""
    key1_data = load_binary(key1)
    key2_data = load_binary(key2)
    first = reconstruct_key(key1_data)
    second = reconstruct_key(key2_data)

    if not (isinstance(first, ECC.EccKey) and isinstance(second, ECC.EccKey)):
        raise SPSDKError("Currently on ECC keys are supported.")

    if first.pointQ == second.pointQ:
        click.echo("Keys match.")
    else:
        raise SPSDKAppError("Keys are NOT a valid pair!")


def get_curve_name(key_length: int) -> str:
    """Get curve name for Crypto library."""
    if key_length <= 32:
        return "p256"
    if key_length <= 48:
        return "p384"
    if key_length in [64, 96]:
        return f"p{key_length * 4}"
    raise SPSDKError(f"Not sure what curve corresponds to {key_length} data")


# we use Crypto instead of cryptography because of binary private key reconstruction
def reconstruct_key(key_data: bytes) -> Union[ECC.EccKey, RSA.RsaKey]:
    """Reconstruct Crypto key from PEM,DER or RAW data."""
    try:
        return RSA.import_key(key_data)
    except (ValueError, TypeError):
        pass
    try:
        return ECC.import_key(key_data)
    except ValueError:
        pass
    # attempt to reconstruct key from raw data
    key_length = len(key_data)
    curve = get_curve_name(key_length)
    # everything under 49 bytes is a private key
    if key_length <= 48:
        # pylint: disable=invalid-name   # 'd' is regular name for private key number
        d = int.from_bytes(key_data, byteorder="big")
        return ECC.construct(curve=curve, d=d)
    # public keys in binary form have exact sizes
    if key_length in [64, 96]:
        coord_length = key_length // 2
        x = int.from_bytes(key_data[:coord_length], byteorder="big")
        y = int.from_bytes(key_data[coord_length:], byteorder="big")
        return ECC.construct(curve=curve, point_x=x, point_y=y)
    raise SPSDKError(f"Can't recognize key with length {key_length}")


@catch_spsdk_error
def safe_main() -> None:
    """Call the main function."""
    sys.exit(main())  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    safe_main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module for DebugMailbox Debug probes support."""

import contextlib
import logging
from typing import Dict, Iterator, Optional, Type

import colorama
import prettytable

from spsdk import SPSDKError
from spsdk.debuggers.debug_probe import DebugProbe, SPSDKDebugProbeError, SPSDKProbeNotFoundError
from spsdk.debuggers.debug_probe_jlink import DebugProbePyLink

# Import all supported debug probe classes
from spsdk.debuggers.debug_probe_pyocd import DebugProbePyOCD

PROBES: Dict[str, Type[DebugProbe]] = {
    "pyocd": DebugProbePyOCD,
    "jlink": DebugProbePyLink,
}

logger = logging.getLogger(__name__)


class ProbeDescription:
    """NamedTuple for DAT record of debug probe description."""

    def __init__(
        self,
        interface: str,
        hardware_id: str,
        description: str,
        probe: Type[DebugProbe],
    ) -> None:
        """Initialization of Debug probe description class.

        param interface: Probe Interface.
        param hardware_id: Probe Hardware ID(Identification).
        param description: Probe Text description.
        param probe: Probe name of the class.
        """
        self.interface = interface
        self.hardware_id = hardware_id
        self.description = description
        self.probe = probe

    def get_probe(self, options: Optional[Dict] = None) -> DebugProbe:
        """Get instance of probe.

        :param options: The dictionary with options
        :return: Instance of described probe.
        """
        return self.probe(hardware_id=self.hardware_id, options=options)


class DebugProbes(list):
    """Helper class for debug probe selection. This class accepts only ProbeDescription object."""

    def append(self, item: ProbeDescription) -> None:
        """Overriding build-in function by check the type.

        :param item: ProbeDescription item.
        :raises SPSDKError: Invalid input types has been used.
        """
        if isinstance(item, ProbeDescription):
            super().append(item)
        else:
            raise SPSDKError("The list accepts only ProbeDescription object")

    def insert(self, index: int, item: ProbeDescription) -> None:  # type: ignore[override]
        """Overriding build-in function by check the type.

        :param item: ProbeDescription item.
        :param index: Index in list to insert.
        :raises SPSDKError: Invalid input types has been used.
        """
        if isinstance(item, ProbeDescription):
            super().insert(index, item)
        else:
            raise SPSDKError("The list accepts only ProbeDescription object")

    def select_probe(self, silent: bool = False) -> ProbeDescription:
        """Perform Probe selection.

        :param silent: When it True, the functions select the probe if applicable without any prints to log
        :return: The record of selected DebugProbe
        :raises SPSDKProbeNotFoundError: No probe has been founded
        """
        if len(self) == 0:
            raise SPSDKProbeNotFoundError("There is no debug probe connected in system!")

        if not silent or len(self) > 1:  # pragma: no cover
            self.print()

        if len(self) == 1:
            # Automatically gets and use only one option\
            i_selected = 0
        else:  # pragma: no cover
            print("Please choose the debug probe: ", end="")
            i_selected = int(input())
            if i_selected > len(self) - 1:
                raise SPSDKProbeNotFoundError("The chosen probe index is out of range")

        return self[i_selected]

    def print(self) -> None:
        """Prints the List of Probes to nice colored table."""
        # Print all PyOCD probes and then Pemicro with local index
        table = prettytable.PrettyTable(["#", "Interface", "Id", "Description"])
        table.align = "l"
        table.header = True
        table.border = True
        table.hrules = prettytable.HEADER
        table.vrules = prettytable.NONE
        i = 0
        for probe in self:
            table.add_row(
                [
                    colorama.Fore.YELLOW + str(i),
                    colorama.Fore.WHITE + probe.interface,
                    colorama.Fore.CYAN + probe.hardware_id,
                    colorama.Fore.GREEN + probe.description,
                ]
            )
            i += 1
        print(table.get_string() + colorama.Style.RESET_ALL)


class DebugProbeUtils:
    """The SPSDK debug probes utilities class.

    The SPSDK debug probes utilities, that helps user to find and open the real
    hardware debug probe to establish connection with hardware.
    """

    @staticmethod
    def get_connected_probes(
        interface: Optional[str] = None,
        hardware_id: Optional[str] = None,
        options: Optional[Dict] = None,
    ) -> DebugProbes:
        """Functions returns the list of all connected probes in system.

        The caller could restrict the scanned interfaces by specification of hardware ID.

        :param interface: None to scan all interfaces, otherwise the selected interface is scanned only.
        :param hardware_id: None to list all probes, otherwise the the only probe with matching
        :param options: The dictionary with optional options
            hardware id is listed.
        :return: list of probe_description's
        """
        probes = DebugProbes()
        for key, probe in PROBES.items():
            if (interface is None) or (interface.lower() == key):
                try:
                    probes.extend(probe.get_connected_probes(hardware_id, options))
                except SPSDKDebugProbeError as exc:
                    logger.warning(f"The {key} debug probe support is not ready({str(exc)}).")

        return probes


def test_ahb_access(probe: DebugProbe, ap_mem: Optional[int] = None, invasive: bool = True) -> bool:
    """The function safely test the access of debug probe to AHB in target.

    :param probe: Probe object to use for test.
    :param ap_mem: Index of memory access port, defaults to 0
    :param invasive: Invasive type of test (temporary changed destination RAM value)
    :return: True is access to AHB is granted, False otherwise.
    """
    ahb_enabled = False
    bck_mem_ap = probe.mem_ap_ix
    probe.mem_ap_ix = ap_mem or probe.mem_ap_ix
    try:
        test_value = probe.mem_reg_read(probe.TEST_MEM_AP_ADDRESS)
        logger.debug(
            f"Test Connection: Read value at {hex(probe.TEST_MEM_AP_ADDRESS)} is {test_value:08X}"
        )
        if invasive:
            probe.mem_reg_write(addr=probe.TEST_MEM_AP_ADDRESS, data=test_value ^ 0xAAAAAAAA)
            test_read = probe.mem_reg_read(probe.TEST_MEM_AP_ADDRESS)
            probe.mem_reg_write(addr=probe.TEST_MEM_AP_ADDRESS, data=test_value)
            if test_read != test_value ^ 0xAAAAAAAA:
                raise SPSDKError("Test connection verification failed")
        ahb_enabled = True

    except SPSDKError as exc:
        logger.debug(f"Test Connection: Chip has NOT enabled AHB access. {str(exc)}")
    finally:
        probe.mem_ap_ix = bck_mem_ap

    return ahb_enabled


@contextlib.contextmanager
def open_debug_probe(
    interface: Optional[str] = None,
    serial_no: Optional[str] = None,
    debug_probe_params: Optional[Dict] = None,
) -> Iterator[DebugProbe]:
    """Method opens DebugProbe object based on input arguments.

    :param interface: None to scan all interfaces, otherwise the selected interface is scanned only.
    :param serial_no: None to list all probes, otherwise the the only probe with matching
    :param debug_probe_params: The dictionary with optional options
        hardware id is listed.
    :return: Active DebugProbe object.
    :raises SPSDKError: Raised with any kind of problems with debug probe.
    """
    debug_probes = DebugProbeUtils.get_connected_probes(
        interface=interface, hardware_id=serial_no, options=debug_probe_params
    )
    selected_probe = debug_probes.select_probe()
    debug_probe = selected_probe.get_probe(debug_probe_params)
    debug_probe.open()

    try:
        yield debug_probe
    except SPSDKError as exc:
        raise SPSDKError("Problem with debug probe occurred") from exc
    finally:
        debug_probe.close()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""This module contains XMCD (External Memory Configuration Data) related code."""


import datetime
import logging
import os
from copy import deepcopy
from typing import Any, Dict, List, Optional

from crcmod.predefined import mkPredefinedCrcFun
from ruamel.yaml.comments import CommentedMap as CM

from spsdk import version as spsdk_version
from spsdk.exceptions import SPSDKError, SPSDKKeyError, SPSDKValueError
from spsdk.image.segments import XMCDHeader
from spsdk.image.segments_base import SegmentBase
from spsdk.image.xmcd import XMCD_DATA_FOLDER, XMCD_DATABASE_FILE, XMCD_SCH_FILE
from spsdk.utils.database import Database
from spsdk.utils.easy_enum import Enum
from spsdk.utils.registers import Registers
from spsdk.utils.schema_validator import ConfigTemplate, ValidationSchemas, check_config

logger = logging.getLogger(__name__)


class MemoryType(Enum):
    """Support memory types Enum."""

    FLEXSPI_RAM = (0, "flexspi_ram", "FlexSPI RAM")
    SEMC_SDRAM = (1, "semc_sdram", "SEMC SDRAM")


class ConfigurationBlockType(Enum):
    """Support configuration blocks Enum."""

    SIMPLIFIED = (0, "simplified", "Simplified configuration")
    FULL = (1, "full", "Full configuration")


class XMCD(SegmentBase):
    """XMCD (External Memory Configuration Data)."""

    def __init__(self, family: str, revision: str = "latest") -> None:
        """XMCD Constructor.

        :param family: Chip family.
        :param config_block_type: Configuration block type: simplified | full.
        :param revision: Optional Chip family revision.
        :raises SPSDKValueError: Unsupported family.
        """
        super().__init__(family, revision)
        self._mem_type: Optional[str] = None
        self._config_type: Optional[str] = None
        self._registers: Registers = Registers(family, base_endianness="little")

    @property
    def mem_type(self) -> str:
        """Memory type."""
        return self._mem_type or "Unknown"

    @mem_type.setter
    def mem_type(self, value: str) -> None:
        """Memory type setter.

        :param value: Value to be set.
        :raises SPSDKValueError: If givem memory type is not supported
        """
        mem_types = XMCD.get_supported_memory_types(self.family, self.revision)
        if value not in mem_types:
            raise SPSDKValueError(f"Unsupported memory type:{value} not in {mem_types}")
        self._mem_type = value

    @property
    def config_type(self) -> str:
        """Configuration type. It can be either simplified or full."""
        return self._config_type or "Unknown"

    @config_type.setter
    def config_type(self, value: str) -> None:
        """Configuration type setter.

        :param value: Value to be set.
        :raises SPSDKValueError: If givem configuration type is not supported
        """
        supported_config_types = XMCD.get_supported_configuration_types(
            self.family, self.mem_type, self.revision
        )
        if value not in supported_config_types:
            raise SPSDKValueError(
                f"Unsupported config type:{value} not in {supported_config_types}"
            )
        self._config_type = value

    @property
    def registers(self) -> Registers:
        """Registers of segment."""
        return self._registers

    @registers.setter
    def registers(self, value: Registers) -> None:
        self._registers = value

    @staticmethod
    def get_database() -> Database:
        """Get the devices database."""
        return Database(XMCD_DATABASE_FILE)

    def parse(self, binary: bytes) -> None:
        """Parse binary block into XMCD object.

        :param binary: binary image.
        :raises SPSDKError: If given binary block size is not equal to block size in header
        """
        header = XMCDHeader.parse(binary[: XMCDHeader.SIZE])
        self.mem_type = MemoryType.name(header.interface)
        self.config_type = ConfigurationBlockType.name(header.block_type)
        option_size = None
        if header.block_size != len(binary):
            raise SPSDKError("Invalid XMCD configuration size")
        if self.mem_type == "flexspi_ram" and self.config_type == "simplified":
            if header.block_size - XMCDHeader.SIZE not in [4, 8]:
                raise SPSDKError("Invalid XMCD configuration size")
            option_size = 0 if len(binary[XMCDHeader.SIZE :]) == 4 else 1
        self.registers = self.load_registers(
            self.family, self.mem_type, self.config_type, self.revision, option_size=option_size
        )
        super().parse(binary)
        crc = self.calculate_crc(binary)
        logger.info(f"CRC value: {crc!r}")

    @staticmethod
    def load_from_config(config: Dict) -> "XMCD":
        """Load configuration file of XMCD.

        :param config: XMCD configuration file.
        :raises SPSDKKeyError: If XMCD settings do not contain required key
        :return: XMCD object.
        """
        check_config(config, XMCD.get_validation_schemas_family())
        family = config["family"]
        revision = config.get("revision", "latest")
        mem_type = config["mem_type"]
        config_type = config["config_type"]
        check_config(config, XMCD.get_validation_schemas(family, mem_type, config_type, revision))

        xmcd_settings: Dict[str, Dict] = config["xmcd_settings"]

        xmcd = XMCD(family=family, revision=revision)
        xmcd.mem_type = mem_type
        xmcd.config_type = config_type

        option_size = None
        if xmcd.mem_type == "flexspi_ram" and xmcd.config_type == "simplified":
            try:
                option_size = xmcd_settings["configOption0"]["bitfields"]["optionSize"]
            except KeyError as exc:
                raise SPSDKKeyError(
                    "XMCD settings schema must contain optionSize in configOption0"
                ) from exc
        xmcd._registers = xmcd.load_registers(
            xmcd.family, xmcd.mem_type, xmcd.config_type, xmcd.revision, option_size=option_size
        )
        xmcd.registers.load_yml_config(xmcd_settings)
        return xmcd

    def calculate_crc(self, data: bytes) -> bytes:
        """Calculate XMCD CRC value.

        :param data: Data to be used for calculation.
        """
        crc = mkPredefinedCrcFun("crc-32-mpeg")(data)
        crc_bytes = crc.to_bytes(4, "little")
        if len(crc_bytes) < 5:
            crc_bytes = crc_bytes.ljust(5, b"\0")
        return crc_bytes

    def create_config(self) -> str:
        """Create current configuration YAML.

        :raises SPSDKError: Registers are not loaded in the object.
        :return: Configuration of XMCD Block.
        """
        config = CM()
        config["family"] = self.family
        config["revision"] = self.revision
        config["mem_type"] = self.mem_type or "Unknown"
        config["config_type"] = self.config_type or "Unknown"
        if len(self.registers.get_registers()) == 0:
            raise SPSDKError(
                "Registers are not loaded. Load the configuration or parse binary first."
            )
        config["xmcd_settings"] = self.registers.create_yml_config()
        config.yaml_set_start_comment(
            f"XMCD configuration for {self.family}.\n"
            f"Created: {datetime.datetime.now().strftime('%d/%m/%Y %H:%M:%S')}.\n"
            f"NXP SPSDK version: {spsdk_version}"
        )
        return ConfigTemplate.convert_cm_to_yaml(config)

    @classmethod
    def get_validation_schemas(
        cls,
        family: str,
        mem_type: str,
        config_type: str,
        revision: str = "latest",
    ) -> List[Dict[str, Any]]:
        """Create the validation schema.

        :param family: Family description.
        :param mem_type: Used memory type.
        :param config_type: Config type: either simplified or full.
        :param revision: Chip revision specification, as default, latest is used.
        :raises SPSDKError: Family or revision is not supported.
        :return: List of validation schemas.
        """
        sch_cfg = deepcopy(ValidationSchemas.get_schema_file(XMCD_SCH_FILE))
        sch_cfg["xmcd_family_rev"]["properties"]["family"]["enum"] = XMCD.get_supported_families()
        revisions = ["latest"]
        revisions.extend(XMCD.get_database().devices.get_by_name(family).revisions.revision_names)
        sch_cfg["xmcd_family_rev"]["properties"]["revision"]["enum"] = revisions
        sch_cfg["xmcd_family_rev"]["properties"]["mem_type"][
            "enum"
        ] = XMCD.get_supported_memory_types(family, revision)
        sch_cfg["xmcd_family_rev"]["properties"]["config_type"][
            "enum"
        ] = XMCD.get_supported_configuration_types(family, mem_type, revision)

        registers = XMCD.load_registers(family, mem_type, config_type, revision)
        sch_cfg["xmcd"]["properties"]["xmcd_settings"] = registers.get_validation_schema()

        return [sch_cfg["xmcd_family_rev"], sch_cfg["xmcd"]]

    @staticmethod
    def get_validation_schemas_family() -> List[Dict[str, Any]]:
        """Create the validation schema just for supported families.

        :return: List of validation schemas for XMCD supported families.
        """
        sch_cfg = deepcopy(ValidationSchemas.get_schema_file(XMCD_SCH_FILE))
        sch_cfg["xmcd_family_rev"]["properties"]["family"]["enum"] = XMCD.get_supported_families()
        return [sch_cfg["xmcd_family_rev"]]

    @staticmethod
    def generate_config_template(
        family: str, mem_type: str, config_type: str, revision: str = "latest"
    ) -> str:
        """Generate configuration for selected family.

        :param family: Family description.
        :param mem_type: Used memory type.
        :param config_type: Config type: either simplified or full.
        :param revision: Chip revision specification, as default, latest is used.
        :return: Template of XMCD Block.
        """
        try:
            schemas = XMCD.get_validation_schemas(family, mem_type, config_type, revision)
        except SPSDKError:
            return ""
        override = {}
        override["family"] = family
        override["revision"] = revision
        override["mem_type"] = mem_type
        override["config_type"] = config_type

        ret = ConfigTemplate(
            f"External Memory Configuration Data template for {family}.",
            schemas,
            override,
        ).export_to_yaml()

        return ret

    @classmethod
    def get_supported_configuration_types(
        cls, family: str, mem_type: str, revision: str = "latest"
    ) -> List[str]:
        """Return list of supported memory interfaces.

        :return: List of supported family interfaces.
        """
        mem_types = cls.get_memory_types(family, revision)
        return list(mem_types[mem_type].keys())

    @staticmethod
    def _load_header_registers(
        family: str, mem_type: str, config_type: str, revision: str
    ) -> Registers:
        """Load header registers of segment.

        :param family: Family description.
        :param mem_type: Used memory type.
        :param config_type: Config type: either simplified or full.
        :param revision: Chip revision specification, as default, latest is used.
        """
        header_file_name = XMCD.get_database().get_device_value("header", family, revision)
        header_file_path = os.path.join(XMCD_DATA_FOLDER, header_file_name)
        regs = Registers(family, base_endianness="little")
        regs.load_registers_from_xml(header_file_path)
        header_reg = regs.find_reg("header")
        block_type_bf = header_reg.find_bitfield("configurationBlockType")
        value = ConfigurationBlockType.get(config_type)
        block_type_bf.set_value(value)
        memory_if_bf = header_reg.find_bitfield("memoryInterface")
        memory_if_bf.set_value(MemoryType.get(mem_type))
        return regs

    @staticmethod
    def _load_block_registers(
        family: str,
        mem_type: str,
        config_type: str,
        revision: str,
        filter_reg: Optional[List[str]] = None,
    ) -> Registers:
        """Load block registers of segment.

        :param family: Family description.
        :param mem_type: Used memory type.
        :param config_type: Config type: either simplified or full.
        :param revision: Chip revision specification, as default, latest is used.
        :raises SPSDKKeyError: Unknown mem_type or config type
        """
        try:
            block_config = XMCD.get_memory_types(family, revision)[mem_type][config_type]
        except KeyError as exc:
            raise SPSDKKeyError(f"Unsupported combination: {mem_type}:{config_type}") from exc
        block_file_path = os.path.join(XMCD_DATA_FOLDER, block_config)
        regs = Registers(family, base_endianness="little")
        regs.load_registers_from_xml(block_file_path, filter_reg=filter_reg)
        return regs

    @staticmethod
    def load_registers(
        family: str,
        mem_type: str,
        config_type: str,
        revision: str,
        option_size: Optional[int] = None,
    ) -> Registers:
        """Load all registers of segment.

        :param family: Family description.
        :param mem_type: Used memory type.
        :param config_type: Config type: either simplified or full.
        :param revision: Chip revision specification, as default, latest is used.
        :param option_size: Number of configuration words to be loaded.

        :raises SPSDKValueError: If option_size has invalid value(only 0 or 1 are allowed)
        """
        regs = XMCD._load_header_registers(family, mem_type, config_type, revision)
        filter_reg = []
        if mem_type == "flexspi_ram" and config_type == "simplified":
            if option_size is not None and option_size > 1:
                raise SPSDKValueError("Option size can be either 0 or 1")
            if option_size == 0:
                filter_reg = ["configOption1"]
        block_regs = XMCD._load_block_registers(
            family, mem_type, config_type, revision, filter_reg=filter_reg
        )
        for block_reg in block_regs.get_registers():
            block_reg.offset += XMCDHeader.SIZE * 8
            regs.add_register(block_reg)
        header_reg = regs.find_reg("header")
        config_block_size = header_reg.find_bitfield("configurationBlockSize")
        image_len = len(regs.image_info())
        config_block_size.set_value(image_len)
        return regs
